/*
 * ntptrace - show the chain from an NTP host leading back to
 *	its source of time
 *
 *	Jeffrey Mogul	DECWRL	13 January 1993
 *
 *	Inspired by a script written by Glenn Trewitt
 *
 *	Large portions stolen from ntpdate.c
 */
#include <stdio.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/signal.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/file.h>
#include <netinet/in.h>
#if __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif
#if defined(HPUX)
#include <utmp.h>
#endif

#include "ntp_fp.h"
#include "ntp.h"
#include "ntp_unixtime.h"
#include "ntptrace.h"

/*
 * Debugging flag
 */
int debug = 0;

int nonames = 0;			/* if set, don't print hostnames */

/*
 * Program name.
 */
char *progname;

/*
 * Systemwide parameters and flags
 */
int sys_retries = 5;			/* # of retry attempts per server */
u_long sys_timeout = 2;			/* timeout time, in seconds */
struct server **sys_servers;		/* the server list */
int sys_numservers = 0;			/* number of servers to poll */
int sys_maxservers = NTP_MAXSTRATUM+1;	/* max number of servers to deal with */
int sys_version = NTP_OLDVERSION;	/* version to poll with */

/*
 * Declarations for time format conversion tables
 */
extern long tstoushi[];
extern long tstousmid[];
extern long tstouslo[];

extern u_long ustotslo[];
extern u_long ustotsmid[];
extern u_long ustotshi[];

extern char *numtoa();
extern char *numtohost();

/*
 * recvbuf lists
 */
struct recvbuf *freelist;	/* free buffers */
struct recvbuf *fulllist;	/* buffers with data */

int full_recvbufs;	/* number of full ones */
int free_recvbufs;

/*
 * File descriptor masks etc. for call to select
 */
int fd;
fd_set fdmask;

/*
 * Miscellaneous flags
 */
int verbose = 0;
int always_step = 0;

extern int errno;

#if __STDC__
void	prmsg(char *fmt, ...);
#else
void	prmsg();
#endif

void printserver();

/*
 * Main program.  Initialize us and loop waiting for I/O and/or
 * timer expiries.
 */
main(argc, argv)
	int argc;
	char *argv[];
{
	struct server *firstserver;
	int errflg;
	int c;
	extern char *optarg;
	extern int optind;
	extern int getopt();
	struct server *addservbyname();
	extern char *Version;
	extern char *emalloc();

	errflg = 0;
	progname = argv[0];

	/*
	 * Decode argument list
	 */
	while ((c = getopt(argc, argv, "do:nr:t:v")) != EOF)
		switch (c) {
		case 'd':
			++debug;
			break;
		case 'n':
			nonames = 1;
			break;
		case 'o':
			sys_version = atoi(optarg);
			break;
		case 'r':
			sys_retries = atoi(optarg);
			if (sys_retries < 1) {
			    (void)fprintf(stderr,
					"%s: retries (%d) too small\n",
					progname, sys_retries);
			    errflg++;
			}
			break;
		case 't':
			sys_timeout = atoi(optarg);
			if (sys_timeout < 1) {
			    (void)fprintf(stderr,
					"%s: timeout (%d) too short\n",
					progname, sys_timeout);
			    errflg++;
			}
			break;
		case 'v':
			verbose = 1;
			break;
		case '?':
			++errflg;
			break;
		default:
			break;
		}
	
	if (errflg || (argc - optind) > 1) {
		(void) fprintf(stderr,
			"usage: %s [-vnd] [-r retries] [-t timeout] [server]\n",
			    progname);
		exit(2);
	}

	sys_servers = (struct server **)
	    emalloc(sys_maxservers * sizeof(struct server *));

	if (debug)
		setlinebuf(stdout);

	if (debug || verbose)
		prmsg("%s", Version);

	if ((argc - optind) == 1)
		firstserver = addservbyname(argv[optind]);
	else
		firstserver = addservbyname("localhost");
		
	if (firstserver == NULL) {
		/* a message has already been printed */
		exit(2);
	}

	/*
	 * Initialize the time of day routines and the I/O subsystem
	 */
	setup_io();
	
	DoTrace(firstserver);

	exit(0);
}

DoTrace(server)
register struct server *server;
{
	int retries = sys_retries;
	extern char *ntoa();
	extern char *ntohost();

	if (!verbose) {
	    if (nonames)
		printf("%s: ", ntoa(&server->srcadr));
	    else
		printf("%s: ", ntohost(&server->srcadr));
	    fflush(stdout);
	}
	while (retries-- > 0) {
	    DoTransmit(server);
	    if (DoReceive(server))
		return;
	}
	if (verbose) {
	    if (nonames)
		printf("%s:\t*Timeout*\n", ntoa(&server->srcadr));
	    else
		printf("%s:\t*Timeout*\n", ntohost(&server->srcadr));
	}
	else
	    printf("\t*Timeout*\n");
}

/*
 * Dotransmit - transmit a packet to the given server
 */
DoTransmit(server)
register struct server *server;
{
	struct pkt xpkt;
	extern char *ntoa();
	void sendpkt();
	void get_systime();

	if (debug)
		printf("DoTransmit(%s)\n", ntoa(&server->srcadr));

	/*
	 * Fill in the packet and let 'er rip.
	 */
	xpkt.li_vn_mode = PKT_LI_VN_MODE(LEAP_NOTINSYNC,
					sys_version, MODE_CLIENT);
	xpkt.stratum = STRATUM_TO_PKT(STRATUM_UNSPEC);
	xpkt.ppoll = NTP_MINPOLL;
	xpkt.precision = NTPTRACE_PRECISION;
	xpkt.rootdelay = htonl(NTPTRACE_DISTANCE);
	xpkt.rootdispersion = htonl(NTPTRACE_DISP);
	xpkt.refid = htonl(NTPTRACE_REFID);
	xpkt.reftime.l_ui = xpkt.reftime.l_uf = 0;
	xpkt.org.l_ui = xpkt.org.l_uf = 0;
	xpkt.rec.l_ui = xpkt.rec.l_uf = 0;

	/*
	 * just timestamp packet and send it away.
	 */
	get_systime(&(server->xmt));
	HTONL_FP(&server->xmt, &xpkt.xmt);
	sendpkt(&(server->srcadr), &xpkt, LEN_PKT_NOMAC);

	if (debug)
		printf("DoTransmit to %s\n", ntoa(&(server->srcadr)));
}

/*
 * DoReceive - attempt to receive a packet from a specific server
 */
DoReceive(server)
register struct server *server;
{
	register int n;
	fd_set fds;
	struct timeval timeout;
	l_fp ts;
	register struct recvbuf *rb;
	int fromlen;
	int status;
	void get_systime();
	void freerecvbuf();

	/*
	 * Loop until we see the packet we want or until we time out
	 */
	for (;;) {
	    fds = fdmask;
	    timeout.tv_sec = sys_timeout;
	    timeout.tv_usec = 0;
	    n = select(fd+1, &fds, (fd_set *)0, (fd_set *)0, &timeout);
	    
	    if (n == 0) {	/* timed out */
		if (debug)
			printf("timeout\n");
		return(0);
	    }
	    else if (n == -1) {
		prmsg("select() error: %m");
		return(0);
	    }
	    get_systime(&ts);
	    
	    if (free_recvbufs == 0) {
		prmsg("no buffers");
		exit(1);
	    }

	    rb = freelist;
	    freelist = rb->next;
	    free_recvbufs--;

	    fromlen = sizeof(struct sockaddr_in);
	    rb->recv_length = recvfrom(fd, (char *)&rb->recv_pkt,
		sizeof(rb->recv_pkt), 0,
		(struct sockaddr *)&rb->srcadr, &fromlen);
	    if (rb->recv_length == -1) {
		    rb->next = freelist;
		    freelist = rb;
		    free_recvbufs++;
		    continue;
	    }

	    /*
	     * Got one.  Mark how and when it got here,
	     * put it on the full list.
	     */
	    rb->recv_time = ts;
	    rb->next = fulllist;
	    fulllist = rb;
	    full_recvbufs++;

	    status = ReceiveBuf(server, rb);

	    freerecvbuf(rb);
	    
	    return(status);
	}
}

/*
 * receive - receive and process an incoming frame
 *	Return 1 on success, 0 on failure
 */
ReceiveBuf(server, rbufp)
	struct server *server;
	struct recvbuf *rbufp;
{
	register struct pkt *rpkt;
	register s_fp di;
	register u_long t10_ui, t10_uf;
	register u_long t23_ui, t23_uf;
	l_fp org;
	l_fp rec;
	l_fp ci;
	struct server *nextserver;
	struct server *addserver();
	extern char *lfptoa(), *fptoa();
	extern char *ntoa();
	struct in_addr nextia;
	

	if (debug) {
		printf("ReceiveBuf(%s, ", ntoa(&server->srcadr));
		printf("%s)\n", ntoa(&rbufp->srcadr));
	}

	/*
	 * Check to see if the packet basically looks like something
	 * intended for us.
	 */
	if (rbufp->recv_length < LEN_PKT_NOMAC) {
		if (debug)
			printf("receive: packet length %d\n",
			    rbufp->recv_length);
		return(0);		/* funny length packet */
	}
	if (rbufp->srcadr.sin_addr.s_addr != server->srcadr.sin_addr.s_addr) {
		if (debug)
			printf("receive: wrong server\n");
		return(0);		/* funny length packet */
	}

	rpkt = &(rbufp->recv_pkt);

	if (PKT_VERSION(rpkt->li_vn_mode) < NTP_OLDVERSION) {
	    if (debug)
		printf("receive: version %d\n", PKT_VERSION(rpkt->li_vn_mode));
	    return(0);
	}
	if (PKT_VERSION(rpkt->li_vn_mode) > NTP_VERSION) {
	    if (debug)
		printf("receive: version %d\n", PKT_VERSION(rpkt->li_vn_mode));
	    return(0);
	}

	if ((PKT_MODE(rpkt->li_vn_mode) != MODE_SERVER
	    && PKT_MODE(rpkt->li_vn_mode) != MODE_PASSIVE)
	    || rpkt->stratum > NTP_MAXSTRATUM) {
		if (debug)
			printf("receive: mode %d stratum %d\n",
			    PKT_MODE(rpkt->li_vn_mode), rpkt->stratum);
		return(0);
	}
	
	/*
	 * Decode the org timestamp and make sure we're getting a response
	 * to our last request.
	 */
	NTOHL_FP(&rpkt->org, &org);
	if (!L_ISEQU(&org, &server->xmt)) {
		if (debug)
			printf("receive: pkt.org and peer.xmt differ\n");
		return(0);
	}
	
	/*
	 * Looks good.  Record info from the packet.
	 */
	server->leap = PKT_LEAP(rpkt->li_vn_mode);
	server->stratum = PKT_TO_STRATUM(rpkt->stratum);
	server->precision = rpkt->precision;
	server->rootdelay = ntohl(rpkt->rootdelay);
	server->rootdispersion = ntohl(rpkt->rootdispersion);
	server->refid = rpkt->refid;
	NTOHL_FP(&rpkt->reftime, &server->reftime);
	NTOHL_FP(&rpkt->rec, &rec);
	NTOHL_FP(&rpkt->xmt, &server->org);

	/*
	 * Make sure the server is at least somewhat sane.  If not, try
	 * again.
	 */
	if ((rec.l_ui == 0 && rec.l_uf == 0) || !L_ISHIS(&server->org, &rec)) {
		return(0);
	}

	/*
	 * Calculate the round trip delay (di) and the clock offset (ci).
	 * We use the equations (reordered from those in the spec):
	 *
	 * d = (t2 - t3) - (t1 - t0)
	 * c = ((t2 - t3) + (t1 - t0)) / 2
	 */
	t10_ui = server->org.l_ui;	/* pkt.xmt == t1 */
	t10_uf = server->org.l_uf;
	M_SUB(t10_ui, t10_uf, rbufp->recv_time.l_ui,
	    rbufp->recv_time.l_uf);	/* recv_time == t0*/

	t23_ui = rec.l_ui;	/* pkt.rec == t2 */
	t23_uf = rec.l_uf;
	M_SUB(t23_ui, t23_uf, org.l_ui, org.l_uf);	/* pkt->org == t3 */

	/* now have (t2 - t3) and (t0 - t1).  Calculate (ci) and (di) */
	ci.l_ui = t10_ui;
	ci.l_uf = t10_uf;
	M_ADD(ci.l_ui, ci.l_uf, t23_ui, t23_uf);
	M_RSHIFT(ci.l_i, ci.l_uf);

	/*
	 * Calculate di in t23 in full precision, then truncate
	 * to an s_fp.
	 */
	M_SUB(t23_ui, t23_uf, t10_ui, t10_uf);
	di = MFPTOFP(t23_ui, t23_uf);

	server->offset = ci;
	server->delay = di;

	printserver(server, stdout);

	/* End of recursion if we reach stratum 1 */
	if (server->stratum <= 1)
		return(1);

	nextia.s_addr = server->refid;
	nextserver = addserver(&nextia);
	DoTrace(nextserver);
	return(1);
}

/*
 * addserver - Allocate a new structure for server.
 *		Returns a pointer to that structure.
 */
struct server *
addserver(iap)
struct in_addr *iap;
{
	register struct server *server;
	static int toomany = 0;
	extern char *emalloc();

	if (sys_numservers >= sys_maxservers) {
		if (!toomany) {
			toomany = 1;
			prmsg(
		"too many servers (> %d) specified, remainder not used",
			    sys_maxservers);
		}
		return(NULL);
	}

	server = (struct server *)emalloc(sizeof(struct server));
	bzero((char *)server, sizeof(struct server));

	server->srcadr.sin_family = AF_INET;
	server->srcadr.sin_addr = *iap;
	server->srcadr.sin_port = htons(NTP_PORT);

	sys_servers[sys_numservers++] = server;
	
	return(server);
}
/*
 * addservbyname - determine a server's address and allocate a new structure
 *	       for it.  Returns a pointer to that structure.
 */
struct server *
addservbyname(serv)
	char *serv;
{
	u_long ipaddr;
	struct in_addr ia;

	if (!getipaddr(serv, &ipaddr)) {
		prmsg("can't find host %s\n", serv);
		return(NULL);
	}

	ia.s_addr = ipaddr;
	return(addserver(&ia));
}

/*
 * Data related to I/O support
 */

#ifndef FD_SET
#define	NFDBITS		32
#define	FD_SETSIZE	32
#define	FD_SET(n, p)	((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define	FD_CLR(n, p)	((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define	FD_ISSET(n, p)	((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#define FD_ZERO(p)	bzero((char *)(p), sizeof(*(p)))
#endif

/*
 * setup_io - initialize I/O data and open socket
 */
setup_io()
{
	register int i;
	register struct recvbuf *rb;
	extern char *emalloc();

	/*
	 * Init buffer free list and stat counters
	 */
	rb = (struct recvbuf *)
	    emalloc((sys_maxservers + 2) * sizeof(struct recvbuf));
	freelist = 0;
	for (i = sys_maxservers + 2; i > 0; i--) {
		rb->next = freelist;
		freelist = rb;
		rb++;
	}

	fulllist = 0;
	full_recvbufs = 0;
	free_recvbufs = sys_maxservers + 2;

	/* create a datagram (UDP) socket */
	if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
		prmsg("socket() failed: %m");
		exit(1);
		/*NOTREACHED*/
	}

	FD_ZERO(&fdmask);
	FD_SET(fd, &fdmask);
}

/*
 * freerecvbuf - make a single recvbuf available for reuse
 */
void
freerecvbuf(rb)
	struct recvbuf *rb;
{
	rb->next = freelist;
	freelist = rb;
	free_recvbufs++;
}


/*
 * sendpkt - send a packet to the specified destination
 */
void
sendpkt(dest, pkt, len)
	struct sockaddr_in *dest;
	struct pkt *pkt;
	int len;
{
	int cc;
	extern char *ntoa();

	cc = sendto(fd, (char *)pkt, len, 0, (struct sockaddr *)dest,
	    sizeof(struct sockaddr_in));
	if (cc == -1) {
		if (errno != EWOULDBLOCK && errno != ENOBUFS)
			prmsg("sendto(%s): %m", ntoa(dest));
	}
}

/*
 * get_systime implements an interface between the (more or less) system
 * independent bits of ntptrace and the peculiarities of dealing with the
 * Unix system clock.
 *
 * get_systime - return the system time in timestamp format
 */
void
get_systime(ts)
	l_fp *ts;
{
	struct timeval tv;

	/*
	 * Quickly get the time of day and convert it
	 */
	(void) gettimeofday(&tv, (struct timezone *)NULL);
	TVTOTS(&tv, ts);
	ts->l_uf += TS_ROUNDBIT;	/* guaranteed not to overflow */
	ts->l_ui += JAN_1970;
	ts->l_uf &= TS_MASK;
}

/*
 * prmsg - prints various error/warning messages
 *
 * Converted to use varargs, much better ... jks
 */
void
#if __STDC__
prmsg(char *fmt, ...)
#else
/*VARARGS*/
prmsg(va_alist)
	va_dcl
#endif
{
#if !__STDC__
	char *fmt;
#endif
	va_list ap;
	char buf[1025], nfmt[256], xerr[50], *err;
	register int c, l;
	register char *n, *f;
	extern int sys_nerr;
	extern char *sys_errlist[];
	int olderrno;

#if __STDC__
	va_start(ap, fmt);
#else
	va_start(ap);
	fmt = va_arg(ap, char *);
#endif

	olderrno = errno;
	n = nfmt;
	f = fmt;
	while ((c = *f++) != '\0' && c != '\n' && n < &nfmt[252]) {
		if (c != '%') {
			*n++ = c;
			continue;
		}
		if ((c = *f++) != 'm') {
			*n++ = '%';
			*n++ = c;
			continue;
		}
		if ((unsigned)olderrno > sys_nerr)
			sprintf((err = xerr), "error %d", olderrno);
		else
			err = sys_errlist[olderrno];
		if (n + (l = strlen(err)) < &nfmt[254]) {
			strcpy(n, err);
			n += strlen(err);
		}
	}
	*n++ = '\n';
	*n = '\0';

	vsprintf(buf, nfmt, ap);

	(void) fprintf(stderr, "%s: %s", progname, buf);

	va_end(ap);
}

/*
 * getipaddr - given a host name, return its host address
 */
int
getipaddr(host, num)
	char *host;
	u_long *num;
{
	struct hostent *hp;
	int decodeipaddr();

	if (decodeipaddr(host, num)) {
		return 1;
	} else if ((hp = gethostbyname(host)) != 0) {
		bcopy(hp->h_addr, (char *)num, sizeof(u_long));
		return 1;
	}
	return 0;
}

/*
 * decodeipaddr - return a host address (this is crude, but careful)
 */
int
decodeipaddr(num, ipaddr)
	char *num;
	u_long *ipaddr;
{
	register char *cp;
	register char *bp;
	register int i;
	register int temp;
	char buf[80];		/* will core dump on really stupid stuff */

	cp = num;
	*ipaddr = 0;
	for (i = 0; i < 4; i++) {
		bp = buf;
		while (isdigit(*cp))
			*bp++ = *cp++;
		if (bp == buf)
			break;

		if (i < 3) {
			if (*cp++ != '.')
				break;
		} else if (*cp != '\0')
			break;

		*bp = '\0';
		temp = atoi(buf);
		if (temp > 255)
			break;
		*ipaddr <<= 8;
		*ipaddr += temp;
	}
	
	if (i < 4)
		return 0;
	*ipaddr = htonl(*ipaddr);
	return 1;
}


/*
 * printserver - print detail information for a server
 */
void
printserver(pp, fp)
	register struct server *pp;
	FILE *fp;
{
	register int i;
	extern char *prettydate();
	extern char *ntoa();
	extern char *ufptoa();
	extern char *lfptoa();
	void printrefid();
	u_fp synchdist;

	synchdist = pp->rootdispersion + (pp->rootdelay/2);

	if (!verbose) {
	    (void) fprintf(fp, "stratum %d, offset %s, synch distance %s",
			   pp->stratum,
			   lfptoa(&pp->offset, 7),
			   ufptoa(synchdist, 7));
	    if (pp->stratum == 1) {
		(void) fprintf(fp, ", refid ");
		printrefid(fp, pp);
	    }
	    (void) fprintf(fp, "\n");
	    return;
	}

	(void) fprintf(fp, "server %s, port %d\n",
	    ntoa(&pp->srcadr), ntohs(pp->srcadr.sin_port));

	(void) fprintf(fp, "stratum %d, precision %d, leap %c%c\n",
	    pp->stratum, pp->precision,
	    pp->leap & 0x2 ? '1' : '0',
	    pp->leap & 0x1 ? '1' : '0');
	
	(void) fprintf(fp, "refid ");
	printrefid(fp, pp);

	(void) fprintf(fp,
	    " delay %s, dispersion %s ",
	    fptoa(pp->delay, 7),
	    ufptoa(pp->dispersion, 4));
	(void) fprintf(fp, "offset %s\n",
	    lfptoa(&pp->offset, 7));
	(void) fprintf(fp, "rootdelay %s, rootdispersion %s",
		fptoa(pp->rootdelay, 7), ufptoa(pp->rootdispersion, 4));
	(void) fprintf(fp, ", synch dist %s\n",
		ufptoa(synchdist, 7));
	
	(void) fprintf(fp, "reference time:      %s\n",
	    prettydate(&pp->reftime));
	(void) fprintf(fp, "originate timestamp: %s\n",
	    prettydate(&pp->org));
	(void) fprintf(fp, "transmit timestamp:  %s\n",
	    prettydate(&pp->xmt));

	(void) fprintf(fp, "\n");

}

void
printrefid(fp, pp)
FILE *fp;
struct server *pp;
{
	char junk[5];
	char *str;

	if (pp->stratum == 1) {
		junk[4] = 0;
		bcopy((char *)&pp->refid, junk, 4);
		str = junk;
		(void) fprintf(fp, "'%s'", str);
	} else {
	    if (nonames) {
		str = numtoa(pp->refid);
		(void) fprintf(fp, "[%s]", str);
	    }
	    else {
		str = numtohost(pp->refid);
		(void) fprintf(fp, "%s", str);
	    }
	}
}

#if defined(NEED_VSPRINTF)
/*
 * This nugget for pre-tahoe 4.3bsd systems
 */
#if !defined(__STDC__) || !__STDC__
#define const
#endif

int
vsprintf(str, fmt, ap)
	char *str;
	const char *fmt;
	va_list ap;
{
	FILE f;
	int len;

	f._flag = _IOWRT+_IOSTRG;
	f._ptr = str;
	f._cnt = 32767;
	len = _doprnt(fmt, ap, &f);
	*f._ptr = 0;
	return (len);
}
#endif

