/*
 * ntp_loopfilter.c - implements the NTP loop filter algorithm
 */

#include <stdio.h>
#include <sys/types.h>
#include <netinet/in.h>

#include "ntp_syslog.h"
#include "ntp_fp.h"
#include "ntp.h"

#if defined(PPS) || defined(PPSDEV) || defined(PPSCD)
#include <sys/file.h>
#include <sys/ioctl.h>
#include <sgtty.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "ntp_refclock.h"
#endif /* PPS || PPSDEV || PPSCD */

#if defined(PPSDEV) || defined(PPSCD)
#ifdef STREAM		/* This requires ttyclk streams driver. */
#include <stropts.h>
#include <sys/clkdefs.h>
#else
/* Line-discipline folks invited to hack here */
/* No special includes needed for CLKLDISC, extra stuff in sys/ioctl.h */
#endif /* STREAM */
#endif /* PPSDEV || PPSCD */

#if defined(PPSCD)
#include "ntp_unixtime.h"
#include <sys/ppsclock.h>
#endif /* PPSCD */
/*
 * The loop filter is implemented in slavish adherence to the
 * specification (Section 5), except that for consistency we
 * mostly carry the quantities in the same units as appendix G.
 *
 * Note that the long values below are the fractional portion of
 * a long fixed-point value.  This limits these values to +-0.5
 * seconds.  When adjustments are capped inside this range (see
 * CLOCK_MAX_{I,F}) both the clock_adjust and the compliance
 * registers should be fine. (When the compliance is above 16, it
 * will at most accumulate 2**CLOCK_MULT times the maximum offset,
 * which means it fits in a s_fp.)
 *
 * The skew compensation is a special case. In version 2, it was
 * kept in ms/4s (i.e., CLOCK_FREQ was 10). In version 3 (Section 5)
 * it seems to be 2**-16ms/4s in a s_fp for a maximum of +-125ppm
 * (stated maximum 100ppm). Since this seems about to change to a
 * larger range, it will be kept in units of 2**-20 (CLOCK_DSCALE)
 * in an s_fp (mainly because that's nearly the same as parts per
 * million). Note that this is ``seconds per second'', whereas a
 * clock adjustment is a 32-bit fraction of a second to be applied
 * every 2**CLOCK_ADJ seconds; to find it, shift the drift right by
 * (CLOCK_DSCALE-16-CLOCK_ADJ). When updating the drift, on the other
 * hand, the CLOCK_FREQ factor from the spec assumes the value to be
 * in ``seconds per 4 seconds''; to get our units, CLOCK_ADJ must be
 * added to the shift.
 */

#define RSH_DRIFT_TO_FRAC (CLOCK_DSCALE - 16)
#define RSH_DRIFT_TO_ADJ (RSH_DRIFT_TO_FRAC - CLOCK_ADJ)
#define RSH_FRAC_TO_FREQ (CLOCK_FREQ + CLOCK_ADJ - RSH_DRIFT_TO_FRAC)

/*
 * Imported from the ntp_proto module
 */
extern u_char sys_stratum;
extern s_fp sys_rootdelay;
extern u_fp sys_rootdispersion;

l_fp last_offset;	/* last adjustment done */
long clock_adjust;	/* clock adjust register (fraction only) */

s_fp drift_comp;	/* drift compensation register */
s_fp max_comp;		/* drift limit imposed by max host clock slew */

int time_constant;	/* log2 of time constant (0 .. 4) */
u_long tcadj_time;	/* last time-constant adjust time */

u_long watchdog_timer;	/* watchdog timer, in seconds */
int first_adjustment;	/* set to 1 if waiting for first adjustment */
int tc_counter;		/* time-constant hold counter */

int pps_control;	/* set to one if pps signal valid */
l_fp pps_delay;		/* pps tuning offset */
int pps_baud;		/* baud rate of PPS line */
u_long pps_update;	/* last pps update time */
int fdpps = 0;		/* pps file descriptor */

#if defined(PPS) || defined(PPSDEV) || defined(PPSCD)
/*
 * This module has support for a 1-pps signal to fine-tune the local
 * clock. The signal is optional; when present and operating within
 * given tolerances in frequency and jitter, it is used to discipline
 * the local clock. In order for this to work, the local clock must be
 * set to within +-500 ms by another means, such as a radio clock or
 * NTP itself. The 1-pps signal is connected via a serial port and
 * gadget box consisting of a one-shot and EIA level-converter. When
 * operated at 38.4 kbps with a SPARCstation IPC, this arrangement has a
 * worst-case jitter less than 26 us. The pps delay configuration
 * declaration can be used to compensate for miscellaneous uart and
 * os delays. Allow about 247 us for uart delays at 38400 bps and
 * -1 ms for SunOS streams nonsense.
 */

#ifndef B19200
#define B19200	EXTA
#endif
#ifndef B38400
#define B38400	EXTB
#endif

/*
 * A really messy way to map an integer baud rate to the system baud rate.
 * There should be a better way.
 */
#define SYS_BAUD(i) \
	( i == 38400 ? B38400 : \
	( i == 19200 ? B19200 : \
	( i == 9600 ? B9600 : \
	( i == 4800 ? B4800 : \
	( i == 2400 ? B2400 : \
	( i == 1200 ? B1200 : \
	( i == 600 ? B600 : \
	( i == 300 ? B300 : 0 ))))))))

#define PPS_BAUD B38400		/* default serial port speed */
#define PPS_MAXAGE 120		/* seconds after which we disbelieve pps */
#define PPS_MAXUPDATE 600	/* seconds after which we disbelieve timecode */
#define	PPS_DEV	"/dev/pps"	/* pps port */
#define PPS_FAC 5		/* pps shift (log2 trimmed samples) */
#define	NPPS 42			/* pps filter size (1<<PPS_FAC+2*PPS_TRIM) */
#define PPS_TRIM 5		/* samples trimmed from median filter */
#define PPS_XCPT "\377"		/* intercept character */

struct refclockio io;		/* given to the I/O handler */
l_fp pps_offset;		/* filtered pps offset */
u_long pps_time;		/* last pps sample time */
u_long nsamples;		/* number of pps samples collected */
long samples[NPPS];		/* median filter for pps samples */
int pps_sample();
/*
 * Imported from ntp_io.c
 */
extern struct interface *loopback_interface;

#endif /* PPS || PPSDEV || PPSCD */

/*
 * Debug flag importation
 */
extern int debug;

/*
 * Imported from timer module
 */
extern u_long current_time;	/* like it says, in seconds */

/*
 * sys_poll and sys_refskew are set here
 */
extern int sys_poll;		/* log2 of system poll interval */
extern l_fp sys_refskew;	/* accumulated skew since last update */
extern u_fp sys_maxd;		/* max dispersion of survivor list */

/*
 * init_loopfilter - initialize loop filter data
 */
void
init_loopfilter()
{
	extern unsigned long tsf_maxslew;
	unsigned long tsf_limit;
#if defined(PPSDEV) || defined(PPSCD)
	void pps_receive();
#if !defined(HPUXGADGET)
	struct sgttyb ttyb;
	extern int io_addclock();
#else
	extern int io_addclock_simple();
#endif /* HPUXGADGET */
#endif /* PPSDEV || PPSCD */
	clock_adjust = 0;
	drift_comp = 0;
	time_constant = 0;
	tcadj_time = 0;
	sys_poll = NTP_MINPOLL;
	watchdog_timer = 0;
	tc_counter = 0;
	last_offset.l_i = 0;
	last_offset.l_f = 0;
	first_adjustment = 1;

/*
 * Limit for drift_comp, minimum of two values. The first is to avoid
 * signed overflow, the second to keep within 75% of the maximum
 * adjustment possible in adj_systime().
 */
	max_comp = 0x7fff0000;
	tsf_limit = ((tsf_maxslew >> 1) + (tsf_maxslew >> 2));
	if ((max_comp >> RSH_DRIFT_TO_ADJ) > tsf_limit)
		max_comp = tsf_limit << RSH_DRIFT_TO_ADJ;

	pps_control = 0;
#if defined(PPSDEV) || defined(PPSCD)
	pps_baud = PPS_BAUD;
	pps_delay.l_i = 0;
	pps_delay.l_f = 0;
	pps_time = pps_update = 0;
	nsamples = 0;
#endif /* PPSDEV || PPSCD */
#if defined(PPSDEV)

	/*
	 * Open pps serial port. We don't care if the serial port comes
	 * up; if not, we  just use the timecode. Therefore, if anything
	 * goes wrong, just reclaim the resources and continue.
	 */
	fdpps = open(PPS_DEV, O_RDONLY);
	if (fdpps == -1) {
		syslog(LOG_ERR,
		    "init_loopfilter: open %s fails", PPS_DEV);
		return;
	}
#if !defined(HPUXGADGET)
        if (ioctl(fdpps, TIOCEXCL, (char *)0) < 0) {
                syslog(LOG_ERR,
		    "init_loopfilter: ioctl(%s, TIOCEXCL) fails", PPS_DEV);
		(void) close(fdpps);
		return;
	}
        ttyb.sg_ispeed = ttyb.sg_ospeed = pps_baud;
#ifdef CLKDISC
	ttyb.sg_erase = ttyb.sg_kill = *PPS_XCPT;
	ttyb.sg_flags = EVENP|ODDP|RAW|CRMOD;
#else
	ttyb.sg_erase = ttyb.sg_kill = 0;
	ttyb.sg_flags = RAW;
#endif /* CLKDISC */
        if (ioctl(fdpps, TIOCSETP, (char *)&ttyb) < 0) {
                syslog(LOG_ERR,
		    "init_loopfilter: ioctl(%s, TIOCSETP) fails", PPS_DEV);
		(void) close(fdpps);
		return;
	}
#endif /* HPUXGADGET */
#if defined(STREAM)
	/*
	 * Pop off existing streams modules and push on clk module. Tickle
	 * the module to timestamp the intercept character.
	 */
	while (ioctl(fdpps, I_POP, 0 ) >= 0) ;
	if (ioctl(fdpps, I_PUSH, "clk" ) < 0) {
		syslog(LOG_ERR,
		    "init_loopfilter: ioctl(%s, I_PUSH, clk) fails", PPS_DEV);
		(void) close(fdpps);
		return;
	}
	if (ioctl(fdpps, CLK_SETSTR, PPS_XCPT) < 0) {
	        syslog(LOG_ERR,
		     "init_loopfilter: ioctl(%s, CLK_SETSTR) fails", PPS_DEV);
		(void) close(fdpps);
		return;
	 }
#endif /* STREAM */
#if  defined(CLKDISC)
	/*
	 * Okay.  Set the line discipline to the clock line discipline, then
	 * give it to the I/O code to start receiving stuff.  Note the
	 * change of line discipline will clear the read buffers, which
	 * makes the change clean if done quickly.
	 */
	{
		unsigned int ldisc;
		ldisc = CLKLDISC;
		if (ioctl(fdpps, TIOCSETD, (char *)&ldisc) < 0) {
			syslog(LOG_ERR,
				"init_loopfilter: ioctl(%s, TIOCSETD): %m", PPS_DEV);
			(void) close(fdpps);
			return;
		}
	}
#endif /* CLKDISC */

	/*
	 * Insert in device list.
	 */
	io.clock_recv = pps_receive;
	io.srcclock = (caddr_t)NULL;
	io.datalen = 0;
	io.fd = fdpps;
#if defined(HPUXGADGET)
	if (!io_addclock_simple(&io)) {
#else
	if (!io_addclock(&io)) {
#endif /* HPUXGADGET */
		syslog(LOG_ERR, "init_loopfilter: addclock %s fails", PPS_DEV);
		(void) close(fdpps);
		return;
	}
#endif /* PPSDEV */
}

/*
 * local_clock - the NTP logical clock loop filter.  Returns 1 if the
 *		 clock was stepped, 0 if it was slewed and -1 if it is
 *		 hopeless.
 */
int
local_clock(fp_offset, from)
	l_fp *fp_offset;		/* best offset estimate */
	struct sockaddr_in *from;	/* who offset is from, for messages */
{
	register long offset;
	register u_long tmp_ui;
	register u_long tmp_uf;
	register long tmp;
	int isneg;
	extern void step_systime();
	extern char *ntoa();
	extern char *lfptoa();
	extern char *mfptoa();
	extern void record_loop_stats();

#ifdef DEBUG
	if (debug > 1)
		printf("local_clock(%s, %s)\n", lfptoa(fp_offset, 9),
		    ntoa(from));
#endif

	/*
	 * Take the absolute value of the offset
	 */
	tmp_ui = fp_offset->l_ui;
	tmp_uf = fp_offset->l_uf;
	if (M_ISNEG(tmp_ui, tmp_uf)) {
		M_NEG(tmp_ui, tmp_uf);
		isneg = 1;
	} else
		isneg = 0;

	/*
	 * If the clock is way off, don't tempt fate by correcting it.
	 */
	if (tmp_ui >= CLOCK_WAYTOOBIG) {
		syslog(LOG_ERR,
		   "Clock appears to be %u seconds %s, something may be wrong",
		    tmp_ui, isneg>0?"fast":"slow");
#ifndef BIGTIMESTEP
		return (-1);
#endif BIGTIMESTEP
	}

	/*
	 * Save this offset for later perusal
	 */
	last_offset = *fp_offset;

	/*
	 * If the magnitude of the offset is greater than CLOCK.MAX, step
	 * the time and reset the registers.
	 */
	if (tmp_ui > CLOCK_MAX_I || (tmp_ui == CLOCK_MAX_I
	    && (u_long)tmp_uf >= (u_long)CLOCK_MAX_F)) {
		if (watchdog_timer < CLOCK_MINSTEP) {
			/* Mustn't step yet, pretend we adjusted. */
			syslog(LOG_INFO,
			       "adjust: STEP dropped (%s offset %s)\n",
			       ntoa(from), lfptoa(fp_offset, 9));
			return (0);
		}
		syslog(LOG_INFO, "adjust: STEP %s offset %s\n",
		    ntoa(from), lfptoa(fp_offset, 9));
		step_systime(fp_offset);

		clock_adjust = 0;
		watchdog_timer = 0;
		first_adjustment = 1;
		pps_update = 0;
		return (1);
	}

	/*
	 * Here we've got an offset small enough to slew.  Note that
	 * since the offset is small we don't have to carry the damned
	 * high order longword in our calculations.
	 *
	 * The time constant and sample interval are approximated with
	 * shifts, as in Section 5 of the v3 spec. The spec procedure
	 * looks strange, as an interval of 64 to 127 seconds seems to
	 * cause multiplication with 128 (and so on). This code lowers
	 * the multiplier by one bit.
	 *
	 * The time constant update goes after adjust and skew updates,
	 * as in appendix G.
	 */
#if defined(PPS) || defined(PPSDEV) || defined(PPSCD)
	/*
	 * If pps samples are valid, update offset, root delay and
	 * root dispersion. This may be a dramatic surprise to high-
	 * stratum clients, since all of a sudden this server looks
	 * like a stratum-1 clock.
	 */
	if (pps_control) {
		last_offset = pps_offset;
		sys_stratum = 1;
		sys_rootdelay = 0;
		sys_rootdispersion = sys_maxd;
	}
#endif /* PPS || PPSDEV || PPSCD */
	offset = last_offset.l_f;
	clock_adjust = offset >> time_constant;

	/*
	 * Calculate the new frequency error. The factor given in the
	 * spec gives the adjustment per 2**CLOCK_ADJ seconds, but we
	 * want it as a (scaled) pure ratio, so we include that factor
	 * now and remove it later.
	 */
	if (first_adjustment) {
		first_adjustment = 0;
	} else {
		/*
		 * Clamp the integration interval to NTP_MAXPOLL.
		 * The bitcounting in Section 5 gives (n+1)-6 for 2**n,
		 * but has a factor 2**6 missing from CLOCK_FREQ.
		 * We make 2**n give n instead. If watchdog_timer is zero,
		 * pretend it's one.
		 */
		tmp = NTP_MAXPOLL;
		tmp_uf = watchdog_timer;
		if (tmp_uf == 0)
			tmp_uf = 1;
		while (tmp_uf < (1<<NTP_MAXPOLL)) {
			tmp--;
			tmp_uf <<= 1;
		}
		/*
		 * We apply the frequency scaling at the same time as
		 * the sample interval; this ensures a safe right-shift.
		 * (as long as it keeps below 31 bits, which current
		 *  parameters should ensure.
		 */
		tmp = (RSH_FRAC_TO_FREQ - tmp) + time_constant + time_constant;
		if (offset < 0)
			tmp = -((-offset) >> tmp);
		else
			tmp = offset >> tmp;
#ifdef DEBUG
		if (debug > 2)
			printf("watchdog %u, gain %u, change %s\n",
			    watchdog_timer, 1<<time_constant,
			    fptoa(tmp, 5));
#endif
		drift_comp += tmp;
#ifdef PPS
                /* under pps control the loop time constants are much shorter
                   and the drift_comp can grow very large very quickly, so
                   clamp it at a reasonable value.
                */
                if(pps_control) {
                    tmp = max_comp >> 1;
                    if (drift_comp >= tmp)
                        drift_comp = tmp;
                    else if (drift_comp <= -tmp)
                        drift_comp = -tmp;
                }
#endif
		/*
		 * Check that the result is in the possible interval
		 */
		if (drift_comp >= max_comp || drift_comp <= -max_comp) {
			syslog(LOG_ERR,
			       "Drift exceeds %s%sppm, cannot cope",
			       drift_comp>0?"+":"-", fptoa(max_comp, 0));
			exit(1);
		}
	}
	watchdog_timer = 0;

	/*
	 * Determine when to adjust the time constant.
	 */
	if (current_time - tcadj_time >= (1 << sys_poll)) {
		tmp = offset;
		if (tmp < 0) tmp = -tmp;
		tmp = tmp >> (16 + CLOCK_WEIGHTTC - time_constant);
		tcadj_time = current_time;
		if (tmp > sys_maxd) {
			tc_counter = 0;
			if (time_constant > 0) time_constant--;
		}
		else {
			tc_counter++;
			if (tc_counter > CLOCK_HOLDTC) {
				tc_counter = 0;
				if (time_constant < CLOCK_MAXTC)
				    time_constant++;
			}
		}
	}
	sys_poll = NTP_MINPOLL + time_constant;

#ifdef DEBUG
	if (debug > 1)
	    printf("adj %s, drft %s, tau %3i\n",
	    	mfptoa((clock_adjust<0?-1:0), clock_adjust, 9),
	    	fptoa(drift_comp, 9), time_constant);
#endif /* DEBUG */

	(void) record_loop_stats(&last_offset, &drift_comp, time_constant);
	
	/*
	 * Whew.  I've had enough.
	 */
	return (0);
}


/*
 * adj_host_clock - Called every 2**CLOCK_ADJ seconds to update host clock
 */
void
adj_host_clock()
{
	register long adjustment;
	extern void adj_systime();
#if defined(PPSCD)
	extern u_long ustotslo[];
	extern u_long ustotsmid[];
	extern u_long ustotshi[];
	struct ppsclockev ev;
	l_fp ts;

	/*
	 * Note that nothing ugly happens even if the CIOGETEV ioctl is not
	 * configured. Correct for signal delays (!) for ultimate finick.
	 */
	if (ioctl(fdpps, CIOGETEV, &ev) >= 0) {
		ts.l_ui = ev.tv.tv_sec + (u_long)JAN_1970;
		TVUTOTSF(ev.tv.tv_usec, ts.l_uf);
		L_SUB(&ts, &pps_delay);
		L_NEG(&ts);
		if (pps_sample(&ts) < 0)
			return;
	}
#endif /* PPSCD */
#if defined(PPS) || defined(PPSDEV) || defined(PPSCD)
	if (pps_time != 0 && current_time - pps_time > PPS_MAXAGE)
	    pps_time = 0;
	if (pps_update != 0 && current_time - pps_update > PPS_MAXUPDATE)
	    pps_update = 0;
	if (pps_time != 0 && pps_update != 0) {
		if (!pps_control)
		    syslog(LOG_INFO, "pps synch");
		pps_control = 1;
	} else {
		if (pps_control)
		    syslog(LOG_INFO, "pps synch lost");
		pps_control = 0;
	}
#endif /* PPS || PPSDEV || PPSCD */
	if (sys_refskew.l_i >= NTP_MAXSKEW)
		sys_refskew.l_f = 0;	/* clamp it */
	else
		L_ADDUF(&sys_refskew, NTP_SKEWINC);
	adjustment = clock_adjust;
	if (adjustment < 0)
		adjustment = -((-adjustment) >> CLOCK_PHASE);
	else
		adjustment >>= CLOCK_PHASE;

	clock_adjust -= adjustment;
	if (drift_comp < 0)
		adjustment -= ((-drift_comp) >> RSH_DRIFT_TO_ADJ);
	else
		adjustment += drift_comp >> RSH_DRIFT_TO_ADJ;
	adj_systime(adjustment);

	watchdog_timer += (1<<CLOCK_ADJ);
	if (watchdog_timer >= NTP_MAXAGE) {
		first_adjustment = 1;	/* don't use next offset for freq */
	}
}


/*
 * loop_config - configure the loop filter
 */
void
loop_config(item, lfp_value, int_value)
	int item;
	l_fp *lfp_value;
	int int_value;		/* This could be a union, but i'm lazy [glh] */
{
	s_fp tmp;
#if defined(PPSDEV) || defined(PPSCD)
#if !defined(HPUXGADGET)
	struct sgttyb ttyb;
#endif /* HPUXGADGET */
#endif /* PPSDEV || PPSCD */

	switch (item) {
	case LOOP_DRIFTCOMP:
		tmp = LFPTOFP(lfp_value);
		if (tmp >= max_comp || tmp <= -max_comp) {
			syslog(LOG_ERR,
			    "loop_config: skew compensation %s too large",
			    fptoa(tmp, 5));
		} else {
			drift_comp = tmp;
		}
		break;
	
	case LOOP_PPSDELAY:
		pps_delay = *lfp_value;
		break;

	case LOOP_PPSBAUD:
#if defined(PPSDEV) || defined(PPSCD)
#if !defined(HPUXGADGET)
		ttyb.sg_ispeed = ttyb.sg_ospeed = SYS_BAUD(int_value);
#ifdef CLKDISC
		ttyb.sg_erase = ttyb.sg_kill = *PPS_XCPT;
		ttyb.sg_flags = EVENP|ODDP|RAW|CRMOD;
#else
		ttyb.sg_erase = ttyb.sg_kill = 0;
		ttyb.sg_flags = RAW;
#endif /* CLKDISC */
		if (ioctl(fdpps, TIOCSETP, (char *)&ttyb) < 0) {
			syslog(LOG_ERR,
			"init_loopfilter: ioctl(%s, TIOCSETP) fails", PPS_DEV);
			return;
		}
		pps_baud = int_value;
#endif /* HPUXGADGET */
#endif /* PPSDEV || PPSCD */
		break;

	default:
		/* sigh */
		break;
	}
}

#if defined(PPSDEV) || defined(PPSCD)

/*
 * pps_receive - compute and store 1-pps signal offset
 * 
 * This routine is called once per second when the 1-pps signal is 
 * present. It calculates the offset of the local clock relative to the
 * 1-pps signal and saves in a circular buffer for later use.
 */
void pps_receive(rbufp)
	struct recvbuf *rbufp;
{
	u_char *dpt;		/* buffer pointer */
	u_char ctmp;		/* intercept char */
	l_fp ts;	        /* l_fp temps */

	/*
	 * Set up pointers, check the buffer length, discard intercept
	 * character and convert unix timeval to timestamp format.
	 */
	dpt = (u_char *)&rbufp->recv_space;
#if defined(HPUXGADGET)
	if (rbufp->recv_length != sizeof(struct timeval)) {
#elif defined(STREAM) || defined(CLKDISC)
        ctmp = *dpt++;
	if (rbufp->recv_length != sizeof(struct timeval) + 1) {
#else
	if (rbufp->recv_length != 1) {
#endif /* HPUXGADGET */
#ifdef DEBUG
		if (debug)
		    printf("pps_receive: bad length %d %c\n",
			rbufp->recv_length, ctmp);
#endif /* DEBUG */
		return;
	}
#if defined(STREAM) || defined(CLKDISC)
	if (!buftvtots(dpt, &ts)) {
#ifdef DEBUG
		if (debug)
		    printf("pps_receive: buftvtots failed\n");
#endif /* DEBUG */
		return;
	}
#else
	ts = rbufp->recv_time;
#endif /* defined(STREAM) || defined(CLKDISC) */

        /*
	 * Correct for uart and os delay and process sample offset.
	 */
	L_SUB(&ts, &pps_delay);
	L_NEG(&ts);
	pps_sample(&ts);
}	
#endif /* PPSDEV || PPSCD */

#if defined(PPS) || defined(PPSDEV) || defined(PPSCD)
/*
 * pps_sample - process pps sample offset
 */
int pps_sample(tsr)
	l_fp *tsr;
{
	int i, j;		/* temp ints */
	long sort[NPPS];	/* temp array for sorting */
	l_fp lftemp, ts;	/* l_fp temps */
	long ltemp;		/* long temp */
	extern void record_peer_stats();
	extern u_short ctlsysstatus();
      
	/*
	 * Note the seconds offset is already in the low-order timestamp
	 * doubleword, so all we have to do is sign-extend and invert it.
	 * The resulting offset is believed only if within CLOCK_MAX.
	 */
	ts = *tsr;
	lftemp.l_i = lftemp.l_f = 0;
	M_ADDF(lftemp.l_i, lftemp.l_f, ts.l_f);
	if (ts.l_f <= -CLOCK_MAX_F || ts.l_f >= CLOCK_MAX_F)
		return (-1);

	/*
	 * Save the sample in a circular buffer for later processing.
	 */
	nsamples++;
	i = ((int)(nsamples)) % NPPS;
	samples[i] = lftemp.l_f;
	if (i != NPPS-1)
	    return (0);

	/*
	 * When the buffer fills up, construct an array of sorted
	 * samples.
	 */
	pps_time = current_time;
	for (i = 0; i < NPPS; i++) {
		sort[i] = samples[i];
		for (j = 0; j < i; j++) {
			if (sort[j] > sort[i]) {
				ltemp = sort[j];
				sort[j] = sort[i];
				sort[i] = ltemp;
			}
		}
	}

	/*
	 * Compute offset as the average of all samples in the filter
	 * less PPS_TRIM samples trimmed from the beginning and end,
	 * dispersion as the difference between max and min of samples
	 * retained. The system stratum, root delay and root dispersion
	 * are also set here.
	 */
	pps_offset.l_i = pps_offset.l_f = 0;
	for (i = PPS_TRIM; i < NPPS - PPS_TRIM; i++)
	    M_ADDF(pps_offset.l_i, pps_offset.l_f, sort[i]);
	if (L_ISNEG(&pps_offset)) {
		L_NEG(&pps_offset);
		for (i = 0; i < PPS_FAC; i++)
		    L_RSHIFT(&pps_offset);
		L_NEG(&pps_offset);
	} else {
		for (i = 0; i < PPS_FAC; i++)
		    L_RSHIFT(&pps_offset);
	}
	lftemp.l_i = 0;
	lftemp.l_f = sort[NPPS-1-PPS_TRIM] - sort[PPS_TRIM];
	sys_maxd = LFPTOFP(&lftemp);
#ifdef DEBUG
	if (debug)
	    printf("pps_filter: %s %s %s\n", lfptoa(&pps_delay, 6),
		lfptoa(&pps_offset, 6), lfptoa(&lftemp, 5));
#endif /* DEBUG */
	record_peer_stats(loopback_interface, ctlsysstatus(), &pps_offset,
	    sys_rootdelay, sys_rootdispersion);
	return (0);
}
#endif /* PPS || PPSDEV || PPSCD */

