/*
 * refclock_gps - clock driver for the Austron 2201 GPS Timing Receiver
 */
#if defined(REFCLOCK) && defined(AS2201)

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/ioctl.h>
#if defined(HPUX) || defined(_AUX_SOURCE)
#include <termio.h>
#else
#include <sgtty.h>
#endif

#include "ntp_syslog.h"
#include "ntp_fp.h"
#include "ntp.h"
#include "ntp_refclock.h"
#include "ntp_unixtime.h"

#ifdef STREAM
#include <stropts.h>
#include <sys/clkdefs.h>
#endif /* STREAM */
#include "ntp_unixtime.h"
#include <sys/ppsclock.h>

/*
 * This driver supports the Austron 2201 GPS Timing Receiver
 *
 * The program sends a "date\r" to the radio and expects a response of
 * the form "hh:mm:ss \r" or "hh:mm:ss*\r", the latter indicating the
 * date has not been acquired. In the reply the nnnnn is the Modified
 * Julian Date. Then, it sends a "time\r" to the radio and expects a
 * response of the form "hh:mm:ss \r" or "hh:mm:ss*\r", the latter
 * indicating the time has not been acquired.
 *
 * This is a work in progress. In order for this code to work, the radio
 * must be placed in non-interactive mode using the "off" command. The
 * radio is not completely reliable; once in a while it fails to respond
 * to a command. The time and date commands are distinct and various
 * inconsistencies are possible near midnight, etc.
 */

/*
 * Definitions
 */
#define	MAXUNITS	4	/* max number of GPS units */
#define	GPS232	"/dev/gps%d"	/* name of radio device */
#define	SPEED232	B9600	/* uart speed (9600 baud) */

/*
 * Radio interface parameters
 */
#define	GPSSKEWFACTOR	0	/* unsynch skew (about 200 ppm) */
#define	GPSPRECISION	(-13)	/* precision assumed (about 100 us) */
#define	GPSREFID	"GPS"	/* reference id */
#define	GPSDESCRIPTION	"Austron 2201 GPS Receiver" /* who we are */
#define	GPSHSREFID	0x7f7f040a /* 127.127.4.10 refid hi strata */
#define GMT		0	/* hour offset from Greenwich */
#define	NCODES		3	/* stages of median filter */
#define	LENDATE		4	/* datecode length */
#define	LENTIME		9	/* timecode length */
#define FMTUNKN		0	/* unknown format timecode id */
#define FMTDATE		1	/* datecode id */
#define FMTTIME		2	/* timecode id */
#define BMAX		100	/* code buffer length */

/*
 * State stuff
 */
#define ST_IDLE		0	/* nothing happening */
#define ST_DATE		1	/* wait for date */
#define ST_TIME		2	/* wait for time */

/*
 * Hack to avoid excercising the multiplier.  I have no pride.
 */
#define	MULBY10(x)	(((x)<<3) + ((x)<<1))

#define	CODEDIFF	0x20000000	/* 0.125 seconds as an l_fp fraction */

/*
 * Time conversion tables imported from the library
 */
extern u_long msutotsflo[];
extern u_long msutotsfhi[];
extern u_long ustotslo[];
extern u_long ustotsmid[];
extern u_long ustotshi[];

/*
 * Imported from ntp_timer module
 */
extern u_long current_time;	/* current time (s) */

/*
 * Imported from ntp_loopfilter module
 */
extern int fdpps;		/* pps file descriptor */

/*
 * Imported from ntpd module
 */
extern int debug;		/* global debug flag */

void as2201_poll();

/*
 * GPS unit control structure.
 */
struct gpsunit {
	struct peer *peer;	/* associated peer structure */
	struct refclockio io;	/* given to the I/O handler */
	struct ppsclockev fv;	/* Unix time of poll */
	l_fp lastrec;		/* last data receive time */
	l_fp lastref;		/* last timecode time */
	l_fp offset[NCODES];	/* recent sample offsets */
	char lastcode[BMAX];	/* last timecode received */
	u_char format;		/* timecode format */
	u_char lencode;		/* length of last timecode */
	u_long lasttime;	/* last time clock heard from */
	u_char unit;		/* unit number for this guy */
	u_char status;		/* clock status */
	u_char lastevent;	/* last clock event */
	u_char reason;		/* reason for last abort */
	u_char year;		/* year of eternity */
	u_short day;		/* day of year */
	u_char hour;		/* hour of day */
	u_char minute;		/* minute of hour */
	u_char second;		/* seconds of minute */
	u_char leap;		/* leap indicators */
	u_long yearstart;	/* start of current year */
	int linect;		/* count of lines remaining */
	int state;		/* current state */
	/*
	 * Status tallies
 	 */
	u_long polls;		/* polls sent */
	u_long noreply;		/* no replies to polls */
	u_long coderecv;	/* timecodes received */
	u_long badformat;	/* bad format */
	u_long baddata;		/* bad data */
	u_long timestarted;	/* time we started this */
};

/*
 * Data space for the unit structures.  Note that we allocate these on
 * the fly, but never give them back.
 */
static struct gpsunit *gpsunits[MAXUNITS];
static u_char unitinuse[MAXUNITS];

/*
 * Keep the fudge factors separately so they can be set even
 * when no clock is configured.
 */
static l_fp fudgefactor[MAXUNITS];
static u_char stratumtouse[MAXUNITS];
static u_char sloppyclockflag[MAXUNITS];

/*
 * as2201_init - initialize internal gps driver data
 */
void
as2201_init()
{
	register int i;
	/*
	 * Just zero the data arrays
	 */
	bzero((char *)gpsunits, sizeof gpsunits);
	bzero((char *)unitinuse, sizeof unitinuse);

	/*
	 * Initialize fudge factors to default.
	 */
	for (i = 0; i < MAXUNITS; i++) {
		fudgefactor[i].l_ui = 0;
		fudgefactor[i].l_uf = 0;
		stratumtouse[i] = 0;
		sloppyclockflag[i] = 0;
	}
}


/*
 * as2201_start - open the GPS devices and initialize data for processing
 */
int
as2201_start(unit, peer)
	u_int unit;
	struct peer *peer;
{
	register struct gpsunit *gps;
	register int i;
	int fd232;
	char gpsdev[20];
	unsigned int ldisc;
#if defined(HPUX) || defined(_AUX_SOURCE)
	struct termio ttyb;
#else
	struct sgttyb ttyb;
#endif
	void as2201_receive();
	extern int io_addclock();
	extern void io_closeclock();
	extern char *emalloc();

	/*
	 * Check configuration info.
	 */
	if (unit >= MAXUNITS) {
		syslog(LOG_ERR,
		    "gps clock: unit number %d invalid (max 3)", unit);
		return (0);
	}
	if (unitinuse[unit]) {
		syslog(LOG_ERR,
		    "gps clock: unit number %d in use", unit);
		return (0);
	}

	/*
	 * Open serial port.
	 */
	(void) sprintf(gpsdev, GPS232, unit);
	fd232 = open(gpsdev, O_RDWR, 0777);
	if (fd232 == -1) {
		syslog(LOG_ERR,
		    "gps clock: open of %s failed: %m", gpsdev);
		return (0);
	}

#if defined(GPSPPS)
	/*
	 * The following code is used only with the 1-pps signal
	 * support. Push the ppsclock module on top of everything
	 * already there. If this fails, not to worry; just log it
	 * and forget it.
	 */
	if (ioctl(fd232, I_PUSH, "ppsclock") < 0) {
		syslog(LOG_ERR,
		    "gps clock: ioctl(%s, I_PUSH, ppsclock) fails", gpsdev);
		fdpps = fd232;
	}
#endif /* GPSPPS */

#if defined(HPUX) || defined(_AUX_SOURCE)
	/*
	 * Don't have exclusive use, but we can do cooked mode and baud
	 * rate.
	 */
	if (ioctl(fd232, TCGETA, (char *)&ttyb) < 0) {
                syslog(LOG_ERR, "gps clock: ioctl(%s, TCGETA): %m", gpsdev);
                goto screwed;
        }
        ttyb.c_iflag = (IGNBRK|ICRNL);
        ttyb.c_oflag = (OPOST|ONLCR);
        ttyb.c_cflag = (SPEED232|CS8|CLOCAL|CREAD|CSTOPB);
        ttyb.c_lflag = ICANON;
	ttyb.c_cc[VEOF] = '\0';
	ttyb.c_cc[VERASE] = '\0';
	ttyb.c_cc[VKILL] = '\0';
        if (ioctl(fd232, TCSETA, (char *)&ttyb) < 0) {
                syslog(LOG_ERR, "gps clock: ioctl(%s, TCSETA): %m", gpsdev);
                goto screwed;
        }
#else
	/*
	 * Set for exclusive use, cooked mode and baud rate.
	 */
	if (ioctl(fd232, TIOCEXCL, (char *)0) < 0) {
		syslog(LOG_ERR,
		    "gps clock: ioctl(%s, TIOCEXCL): %m", gpsdev);
		goto screwed;
	}
	ttyb.sg_ispeed = ttyb.sg_ospeed = SPEED232;
	ttyb.sg_flags = CRMOD|NOHANG;
	if (ioctl(fd232, TIOCSETP, (char *)&ttyb) < 0) {
		syslog(LOG_ERR,
		    "gps clock: ioctl(%s, TIOCSETP): %m", gpsdev);
		goto screwed;
	}
#endif

	/*
	 * Looks like this might succeed.  Find memory for the
	 * structure. Look to see if there are any unused ones, if not
	 * we malloc() one.
	 */
	if (gpsunits[unit] != 0) {
		gps = gpsunits[unit];	/* The one we want is okay */
	} else {
		for (i = 0; i < MAXUNITS; i++) {
			if (!unitinuse[i] && gpsunits[i] != 0)
				break;
		}
		if (i < MAXUNITS) {
			/*
			 * Reclaim this one
			 */
			gps = gpsunits[i];
			gpsunits[i] = 0;
		} else {
			gps = (struct gpsunit *)
			    emalloc(sizeof(struct gpsunit));
		}
	}
	bzero((char *)gps, sizeof(struct gpsunit));
	gpsunits[unit] = gps;

	/*
	 * Set up the structures
	 */
	gps->peer = peer;
	gps->unit = (u_char)unit;
	gps->timestarted = current_time;
	gps->state = ST_IDLE;

	gps->io.clock_recv = as2201_receive;
	gps->io.srcclock = (caddr_t)gps;
	gps->io.datalen = 0;
	gps->io.fd = fd232;

	/*
	 * Flush buffers and link in I/O list.
	 */
#if defined(HPUX) || defined(_AUX_SOURCE)
	if (ioctl(fd232, TCFLSH, 2) < 0) {
		syslog(LOG_ERR, "gps clock: ioctl(%s, TCFLSH): %m", gpsdev);
#else
        ldisc = 0;
	if (ioctl(fd232, TIOCFLUSH, (char *)&ldisc) < 0) {
		syslog(LOG_ERR,
		    "gps clock: ioctl(%s, TIOCFLUSH): %m", gpsdev);
#endif
		goto screwed;
	}
	if (!io_addclock(&gps->io)) {
		goto screwed;
	}

	/*
	 * All done.  Initialize a few random peer variables, then
	 * return success. Note that root delay and root dispersion are
	 * always zero for this clock.
	 */
	peer->precision = GPSPRECISION;
	peer->rootdelay = 0;
	peer->rootdispersion = 0;
	peer->stratum = stratumtouse[unit];
	if (stratumtouse[unit] <= 1)
	    bcopy(GPSREFID, (char *)&peer->refid, 4);
	else
	    peer->refid = htonl(GPSHSREFID);
	unitinuse[unit] = 1;
	return (1);

	/*
	 * Something broke; abandon ship.
	 */
screwed:
	(void) close(fd232);
	return (0);
}

/*
 * as2201_shutdown - shut down a GPS clock
 */
void
as2201_shutdown(unit)
	int unit;
{
	register struct gpsunit *gps;
	extern void io_closeclock();

	if (unit >= MAXUNITS) {
		syslog(LOG_ERR,
		  "gps clock: INTERNAL ERROR, unit number %d invalid (max 4)",
		    unit);
		return;
	}
	if (!unitinuse[unit]) {
		syslog(LOG_ERR,
		 "gps clock: INTERNAL ERROR, unit number %d not in use", unit);
		return;
	}

	/*
	 * Tell the I/O module to turn us off.  We're history.
	 */
	gps = gpsunits[unit];
	io_closeclock(&gps->io);
	unitinuse[unit] = 0;
}


/*
 * as2201_report_event - note the occurance of an event
 *
 * This routine presently just remembers the report and logs it, but
 * does nothing heroic for the trap handler.
 */
void
as2201_report_event(gps, code)
	struct gpsunit *gps;
	int code;
{
	struct peer *peer;

	peer = gps->peer;
	if (gps->status != (u_char)code) {
		gps->status = (u_char)code;
		if (code != CEVNT_NOMINAL)
			gps->lastevent = (u_char)code;
		/*
		 * Should report event to trap handler in here.
		 * Soon...
		 */
	syslog(LOG_INFO, "clock %s event %x\n",
	    ntoa(&peer->srcadr), code);

	}
}


/*
 * as2201_receive - receive data from the serial interface
 */
void
as2201_receive(rbufp)
	struct recvbuf *rbufp;
{
	register int i;
	register struct gpsunit *gps;
	struct ppsclockev ev;
	register u_char *dpt;
	register char *cp;
	register u_char *dpend;
	u_long tmp_ui, tmp_uf;
	l_fp tstmp;
	u_fp dispersion;
	char as2201_process();
	extern void refclock_receive();
	char *ulfptoa();

	/*
	 * Get the clock this applies to and a pointers to the data.
	 * Edit the timecode to remove control chars
	 */
	gps = (struct gpsunit *)rbufp->recv_srcclock;
	dpt = (u_char *)&rbufp->recv_space;
	dpend = dpt + rbufp->recv_length;
	cp = gps->lastcode;
	while (dpt < dpend)
		if ((*cp = 0x7f & *dpt++) >= ' ') cp++; 
	*cp = '\0';
	gps->lencode = cp - gps->lastcode;
	if (gps->lencode == 0) return;
#ifdef DEBUG
	if (debug)
        	printf("gps: timecode %d %s\n",
		    gps->lencode, gps->lastcode);
#endif

	/*
	 * We get down to business, check the timecode format and decode
	 * its contents. If the timecode has invalid length or is not in
	 * proper format, it declares bad format and exits; if the
	 * converted decimal values are out of range, it declares bad
	 * data and exits.
	 */
	cp = gps->lastcode;
	if (gps->linect == 0) {
		gps->linect = atoi(cp);
		return;
		}
	gps->lastrec = rbufp->recv_time;
	gps->linect--;
	gps->leap = 0;
	gps->format = FMTUNKN;
	switch (gps->state) {

	case ST_DATE:

		/*
	 	 * Check date format (doy)
	 	 */
		if (gps->lencode != LENDATE || !isdigit(cp[0]) ||
			!isdigit(cp[1]) || !isdigit(cp[2])) {
				gps->badformat++;
				as2201_report_event(gps, CEVNT_BADREPLY);
				return;
			}
		else gps->format = FMTDATE;

		/*
		 * Convert date and check values 
		 */
		gps->day = cp[0] - '0';
		gps->day = MULBY10(gps->day) + cp[1] - '0';
		gps->day = MULBY10(gps->day) + cp[2] - '0';
		if (cp[3] != ' ')
		    gps->leap = LEAP_NOTINSYNC;
		else
		    gps->lasttime = current_time;
		if (gps->day < 1 || gps->day > 366) {
			gps->baddata++;
			as2201_report_event(gps, CEVNT_BADDATE);
			return;
		}
		gps->state = ST_TIME;
		gps->linect = 0;
		if (write(gps->io.fd, "time\r", 5) != 5) {
			syslog(LOG_ERR,
		   	    "gps clock: write fails to unit %d: %m",
			    gps->unit);
			as2201_report_event(gps, CEVNT_FAULT);
		}
		return;

	case ST_TIME:

		/*
	 	 * Check time format hh:mm:ss
	 	 */
		if (gps->lencode != LENTIME || !isdigit(cp[0]) ||
			!isdigit(cp[1]) || cp[2] != ':' ||
			!isdigit(cp[3]) || !isdigit(cp[4]) ||
			cp[5] != ':' ||	!isdigit(cp[6]) ||
			!isdigit(cp[7])) {
				gps->badformat++;
				as2201_report_event(gps, CEVNT_BADREPLY);
				return;
			}
		else gps->format = FMTTIME;

		/*
		 * Convert time and check values 
		 */
		gps->hour = MULBY10(cp[0] - '0') + cp[1] - '0';
		gps->minute = MULBY10(cp[3] - '0') + cp[4] -  '0';
		gps->second = MULBY10(cp[6] - '0') + cp[7] - '0';
		if (cp[8] != ' ')
		    gps->leap = LEAP_NOTINSYNC;
		else 
		    gps->lasttime = current_time;
		if (gps->hour > 23 || gps->minute > 59
		    || gps->second > 59) {
			gps->baddata++;
			as2201_report_event(gps, CEVNT_BADTIME);
			return;
		}
		gps->state = ST_IDLE;
	}
	/*
	 * Now, compute the reference time value. Use the heavy
	 * machinery for the seconds, which presumably is one which
	 * occured at the last pps pulse, which is captured by the
	 * loop_filter module. All we have to do here is present a
	 * reasonable facsimile of the time at that pulse so the clock-
	 * filter and selection machinery declares us truechimer. The
	 * precision offset within the second is really tuned by the
	 * loop_filter module. Note that this code does not yet know how
	 * to do the years and relies on the clock-calendar chip for
	 * sanity. To admit the radio can honk MJD is embarrassing
	 */
	if (!clocktime(gps->day, gps->hour, gps->minute,
	    gps->second, GMT, gps->lastrec.l_ui,
	    &gps->yearstart, &gps->lastref.l_ui)) {
		gps->baddata++;
		as2201_report_event(gps, CEVNT_BADTIME);
		return;
	}
	gps->lastref.l_uf = 0;
	if (ioctl(fdpps, CIOGETEV, (char *)&ev) < 0) {
		as2201_report_event(gps, CEVNT_FAULT);
		syslog(LOG_ERR,
		    "gps: ioctl CIOGETEV fails for unit %d: %m",
		    gps->unit);
		return;
	}

	/*
	 * If we didn't read the date and time in the same second, try
	 * again.
	 */
	if (gps->fv.tv.tv_sec != ev.tv.tv_sec ||
	    gps->fv.tv.tv_usec != ev.tv.tv_usec) {
		as2201_poll(gps->unit, (char *)0);
		return;
	}
	tstmp.l_i = ev.tv.tv_sec + (u_long)JAN_1970;
	TVUTOTSF(ev.tv.tv_usec, tstmp.l_f);
	L_SUB(&gps->lastref, &tstmp);
	i = ((int)(gps->coderecv)) % NCODES;
	gps->offset[i] = gps->lastref;
	if (gps->coderecv == 0)
		for (i = 1; i < NCODES; i++)
			gps->offset[i] = gps->offset[0];
#if DEBUG
	if (debug)
		printf("gps: times %s %s %s\n", lfptoa(gps->lastrec, 6),
			lfptoa(gps->lastref, 6));
#endif
	gps->coderecv++;

	/*
	 * Process the samples in the median filter, add the fudge
	 * factor and pass the offset and dispersion along. We use
	 * lastref as both the reference time and receive time in order
	 * to avoid being cute, like setting the reference time later
	 * than the receive time, which may cause a paranoid protocol
	 * module to chuck out the data.
 	 */
	if (!as2201_process(gps, &tstmp, &dispersion)) {
		gps->baddata++;
		as2201_report_event(gps, CEVNT_BADTIME);
		return;
	}
	L_ADD(&tstmp, &(fudgefactor[gps->unit]));
	refclock_receive(gps->peer, &tstmp, GMT, dispersion,
	    &gps->lastref, &gps->lastref, gps->leap);
}

/*
 * as2201_process - process a pile of samples from the clock
 *
 * This routine uses a three-stage median filter to calculate offset and
 * dispersion. reduce jitter. The dispersion is calculated as the span
 * of the filter (max - min), unless the quality character (format 2) is
 * non-blank, in which case the dispersion is calculated on the basis of
 * the inherent tolerance of the internal radio oscillator, which is
 * +-2e-5 according to the radio specifications.
 */
char
as2201_process(gps, offset, dispersion)
	struct gpsunit *gps;
	l_fp *offset;
	u_fp *dispersion;
{
	register int i, j;
	register u_long tmp_ui, tmp_uf;
	int not_median1, not_median2, median;
	u_fp disp_tmp, disp_tmp2;

	/*
	 * This code implements a three-stage median filter. First, we
         * check if the samples are within 125 ms of each other. If not,
	 * dump the sample set. We take the median of the three offsets
	 * and use that as the sample offset. There probably is not much
	 * to be gained by a longer filter, since the clock filter in
	 * ntp_proto should do its thing.
	 */
	disp_tmp2 = 0;
	for (i = 0; i < NCODES-1; i++) {
		for (j = i+1; j < NCODES; j++) {
			tmp_ui = gps->offset[i].l_ui;
			tmp_uf = gps->offset[i].l_uf;
			M_SUB(tmp_ui, tmp_uf, gps->offset[j].l_ui,
				gps->offset[j].l_uf);
			if (M_ISNEG(tmp_ui, tmp_uf)) {
				M_NEG(tmp_ui, tmp_uf);
			}
			if (tmp_ui != 0 || tmp_uf > CODEDIFF) {
				return (0);
			}
			disp_tmp = MFPTOFP(0, tmp_uf);
			if (disp_tmp > disp_tmp2) {
				disp_tmp2 = disp_tmp;
				not_median1 = i;
				not_median2 = j;
			}
		}
	}
	if (gps->lasttime == 0)
	    disp_tmp2 = NTP_MAXDISPERSE;
	else
	    disp_tmp2 = current_time - gps->lasttime;
	if (not_median1 == 0) {
		if (not_median2 == 1)
		    median = 2;
		else
		    median = 1;
        } else {
		median = 0;
        }
	*offset = gps->offset[median];
	*dispersion = disp_tmp2;
	return (1);
}

/*
 * as2201_poll - called by the transmit procedure
 */
void
as2201_poll(unit, peer)
	int unit;
	char *peer;
{
	struct gpsunit *gps;

	if (unit >= MAXUNITS) {
		syslog(LOG_ERR,
		    "gps clock poll: INTERNAL: unit %d invalid", unit);
		return;
	}
	if (!unitinuse[unit]) {
		syslog(LOG_ERR,
		    "gps clock poll: INTERNAL: unit %d unused", unit);
		return;
	}
	gps = gpsunits[unit];
	if ((current_time - gps->lasttime) > 150) {
		as2201_report_event(gpsunits[unit], CEVNT_TIMEOUT);
	}
	if (gps->state != ST_IDLE)
		gps->noreply++;

	/*
	 * Earn some revenue. Initialize state and line counters, record
	 * the (Unix) time and send a "date" command.
	 */
	gps->state = ST_DATE;
	gps->linect = 0;
	if (ioctl(fdpps, CIOGETEV, &gps->fv) < 0) {
		as2201_report_event(gps, CEVNT_FAULT);
		syslog(LOG_ERR,
		    "gps: ioctl CIOGETEV fails for unit %d: %m",
		    gps->unit);
		return;
	}
	if (write(gps->io.fd, "date\r", 5) != 5) {
		syslog(LOG_ERR,
		    "gps clock: write fails to unit %d: %m", gps->unit);
		as2201_report_event(gps, CEVNT_FAULT);
	} else {
		gps->polls++;
	}
}

/*
 * as2201_control - set fudge factors, return statistics
 */
void
as2201_control(unit, in, out)
	u_int unit;
	struct refclockstat *in;
	struct refclockstat *out;
{
	register struct gpsunit *gps;

	if (unit >= MAXUNITS) {
		syslog(LOG_ERR,
		    "gps clock: unit %d invalid (max %d)", unit, MAXUNITS-1);
		return;
	}

	if (in != 0) {
		if (in->haveflags & CLK_HAVETIME1)
			fudgefactor[unit] = in->fudgetime1;
		if (in->haveflags & CLK_HAVEVAL1) {
			stratumtouse[unit] = (u_char)(in->fudgeval1 & 0xf);
			if (unitinuse[unit]) {
				struct peer *peer;

				/*
				 * Should actually reselect clock, but
				 * will wait for the next timecode
				 */
				gps = gpsunits[unit];
				peer = gps->peer;
				peer->stratum = stratumtouse[unit];
				if (stratumtouse[unit] <= 1)
					bcopy(GPSREFID, (char *)&peer->refid,
					    4);
				else
					peer->refid = htonl(GPSHSREFID);
			}
		}
		if (in->haveflags & CLK_HAVEFLAG1) {
			sloppyclockflag[unit] = in->flags & CLK_FLAG1;
		}
	}

	if (out != 0) {
		out->type = REFCLK_GPS_AS2201;
		out->haveflags
		    = CLK_HAVETIME1|CLK_HAVEVAL1|CLK_HAVEVAL2|CLK_HAVEFLAG1;
		out->clockdesc = GPSDESCRIPTION;
		out->fudgetime1 = fudgefactor[unit];
		out->fudgetime2.l_ui = 0;
		out->fudgetime2.l_uf = 0;
		out->fudgeval1 = (long)stratumtouse[unit];
		out->fudgeval2 = 0;
		out->flags = sloppyclockflag[unit];
		if (unitinuse[unit]) {
			gps = gpsunits[unit];
			out->lencode = gps->lencode;
			out->lastcode = gps->lastcode;
			out->timereset = current_time - gps->timestarted;
			out->polls = gps->polls;
			out->noresponse = gps->noreply;
			out->badformat = gps->badformat;
			out->baddata = gps->baddata;
			out->lastevent = gps->lastevent;
			out->currentstatus = gps->status;
		} else {
			out->lencode = 0;
			out->lastcode = "";
			out->polls = out->noresponse = 0;
			out->badformat = out->baddata = 0;
			out->timereset = 0;
			out->currentstatus = out->lastevent = CEVNT_NOMINAL;
		}
	}
}

/*
 * as2201_buginfo - return clock dependent debugging info
 */
void
as2201_buginfo(unit, bug)
	int unit;
	register struct refclockbug *bug;
{
	register struct gpsunit *gps;
	register int i;
	register int n;

	if (unit >= MAXUNITS) {
		syslog(LOG_ERR, "gps clock: unit %d invalid (max %d)",
		    unit, MAXUNITS-1);
		return;
	}

	if (!unitinuse[unit])
		return;
	gps = gpsunits[unit];

	bug->nvalues = 11;
	bug->ntimes = 5;
	if (gps->lasttime != 0)
		bug->values[0] = current_time - gps->lasttime;
	else
		bug->values[0] = 0;
	bug->values[1] = (u_long)gps->reason;
	bug->values[2] = (u_long)gps->year;
	bug->values[3] = (u_long)gps->day;
	bug->values[4] = (u_long)gps->hour;
	bug->values[5] = (u_long)gps->minute;
	bug->values[6] = (u_long)gps->second;
	bug->values[8] = gps->noreply;
	bug->values[9] = gps->yearstart;
	bug->stimes = 0x1c;
	bug->times[0] = gps->lastref;
	bug->times[1] = gps->lastrec;
	bug->times[2] = gps->offset[0];
	bug->times[3] = gps->offset[1];
	bug->times[4] = gps->offset[2];
}
#endif
