#if defined(REFCLOCK) && defined(DCF)

/*
 * $Header: /src/NTP/REPOSITORY/v3/xntpd/refclock_dcf77.c,v 3.14 1993/01/19 09:32:48 kardel Exp $
 * 
 * $Id: refclock_dcf77.c,v 3.14 1993/01/19 09:32:48 kardel Exp $
 *
 * reference clock driver for the Meinberg DCF77 clock and others
 * driven by a streams module on top of the tty (SunOS4.x)
 * if the compile time option STREAM is not defined the module
 * will default to standard tty usage. Due to SunOS 4.x STREAMS
 * implementation a jitter of up to 20ms may be observered.
 *
 * Copyright (c) 1989,1990,1991,1992,1993
 * Frank Kardel Friedrich-Alexander Universitaet Erlangen-Nuernberg
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Log: refclock_dcf77.c,v $
 * Revision 3.14  1993/01/19  09:32:48  kardel
 * Release 1993/01/19 DCF77/PPS
 *
 */

#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/termios.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/errno.h>

#include "ntp_syslog.h"
#include "ntp_fp.h"
#include "ntp.h"
#include "ntp_refclock.h"
#include "ntp_unixtime.h"
#include "dcf77.h"

#if !defined(NO_SCCSID) && !defined(lint)
static char rcsid[]="$Id: refclock_dcf77.c,v 3.14 1993/01/19 09:32:48 kardel Exp $";
#endif

extern char *ctime();
extern char *rindex();

/*
 * This driver currently provides the support for
 *   - Meinberg DCF77 receiver DCF77 PZF 535 (TCXO/OCXO versions)
 *   - Meinberg DCF77 receiver U/A 31
 *   - ELV DCF7000
 *
 * Meinberg receivers are connected via a 9600 baud serial line
 * 
 * Some receivers do NOT support:
 *          - announcement of alternate sender usage
 *          - leap second indication
 *
 * so...
 *          - for PZF535 please ask for revision PZFUERL4.6 or higher
 * 
 * Meinberg receiver setup:
 *	output time code every second
 *	Baud rate 9600 7E2S
 */

/*
 * the unit field selects for one the prot to be used (lower 4 bits)
 * and for the other the clock type in case of different but similar
 * receivers (bits 4-6)
 * the most significat bit encodes PPS support
 * when the most significant bit is set the pps telegrams will be used
 * for controlling the local clock (ntp_loopfilter.c)
 * receiver specific configration data is kept in the clockinfo field.
 */

/*
 * Definitions
 */
#define	MAXUNITS	16	/* maximum number of DCF units permitted */
#define DCFDEVICE	"/dev/dcf77-%d" /* device to open %d is unit number */

#define NO_POLL		(void (*)())0
#define NO_INIT		(int  (*)())0
#define NO_END		(void (*)())0
#define NO_DATA		(void *)0
#define NO_FORMAT	""

#define DCF_ID		"DCF"

/*
 * receiver specific constants
 */
#define MBG_CFLAG		(B9600|CS7|PARENB|CREAD|HUPCL)
#define MBG_IFLAG		(IGNBRK|IGNPAR|ISTRIP)
#define MBG_OFLAG		0
#define MBG_LFLAG		0
/*
 * Meinberg DCF U/A 31 (AM) receiver
 */
#define	DCFUA31_PRECISION	(-8)	    /* accounts for receiver errors */
#define	DCFUA31_ROOTDELAY	0x00000D00  /* 50.78125ms */
#define	DCFUA31_BASEDELAY	0x02C00000  /* 10.7421875ms: 10 ms (+/- 3 ms) */
#define	DCFUA31_DESCRIPTION	"Meinberg DCF77 U/A 31"
#define DCFUA31_MAXUNSYNC       60*30       /* only trust clock for 1/2 hour */
#define DCFUA31_CFLAG           MBG_CFLAG
#define DCFUA31_IFLAG           MBG_IFLAG
#define DCFUA31_OFLAG           MBG_OFLAG
#define DCFUA31_LFLAG           MBG_LFLAG

/*
 * Meinberg DCF PZF535/TCXO (FM/PZF) receiver
 */
#define	DCFPZF535_PRECISION	(-18)	    /* 355kHz clock (f(DCF77)*4(PLL))*/
#define	DCFPZF535_ROOTDELAY	0x00000034  /* 800us */
#define	DCFPZF535_BASEDELAY	0x00800000  /* 1.968ms +- 104us (oscilloscope) - relative to start (end of STX) */
#define	DCFPZF535_DESCRIPTION	"Meinberg DCF77 PZF 535/TCXO"
#define DCFPZF535_MAXUNSYNC     60*60*12           /* only trust clock for 12 hours
						    * @ 5e-8df/f we have accumulated
						    * at most 2.16 ms (thus we move to
						    * NTP synchronisation */
#define DCFPZF535_CFLAG         MBG_CFLAG
#define DCFPZF535_IFLAG         MBG_IFLAG
#define DCFPZF535_OFLAG         MBG_OFLAG
#define DCFPZF535_LFLAG         MBG_LFLAG


/*
 * Meinberg DCF PZF535/OCXO receiver
 */
#define	DCFPZF535OCXO_PRECISION	(-18)	   /* 355kHz clock (f(DCF77)*4(PLL))*/
#define	DCFPZF535OCXO_ROOTDELAY	0x00000034 /* 800us (max error * 10) */
#define	DCFPZF535OCXO_BASEDELAY	0x00800000 /* 1.968ms +- 104us (oscilloscope) - relative to start (end of STX) */
#define	DCFPZF535OCXO_DESCRIPTION "Meinberg DCF77 PZF 535/OCXO"
#define DCFPZF535OCXO_MAXUNSYNC     60*60*96       /* only trust clock for 4 days
						    * @ 5e-9df/f we have accumulated
						    * at most an error of 1.73 ms
						    * (thus we move to NTP synchronisation) */
#define DCFPZF535OCXO_CFLAG         MBG_CFLAG
#define DCFPZF535OCXO_IFLAG         MBG_IFLAG
#define DCFPZF535OCXO_OFLAG         MBG_OFLAG
#define DCFPZF535OCXO_LFLAG         MBG_LFLAG


/*
 * ELV DCF7000 Wallclock-Receiver/Switching Clock (Kit)
 *
 * This is really not the hottest clock - but before you have nothing ...
 */
#define DCF7000_PRECISION (-8)	/*
				 * I don't hav much faith in this - but I haven't
				 * evaluated it yet
				 */
#define DCF7000_ROOTDELAY	0x00000364 /* 13 ms */
#define DCF7000_BASEDELAY	0x67AE0000 /* 405 ms - slow blow */
#define DCF7000_DESCRIPTION	"ELV DCF7000"
#define DCF7000_MAXUNSYNC	(60*5) /* sorry - but it just was not build as a clock */
#define DCF7000_CFLAG           (B9600|CS8|CREAD|PARENB|PARODD|CLOCAL|HUPCL)
#define DCF7000_IFLAG		(IGNBRK)
#define DCF7000_OFLAG		0
#define DCF7000_LFLAG		0

struct clockinfo
{
  void (*cl_poll)();		/* active poll routine */
  int  (*cl_init)();		/* active poll init routine */
  void (*cl_end)();		/* active poll end routine */
  void  *cl_data;		/* local data area for "poll" mechanism */
  u_fp   cl_rootdelay;		/* rootdelay */
  u_long cl_basedelay;		/* current offset - unsigned l_fp fractional part */
  s_char cl_precision;		/* device precision */
  char * cl_id;			/* ID code (usually "DCF") */
  char * cl_description;	/* device name */
  char * cl_format;		/* fixed format */
  u_long cl_maxunsync;		/* time to trust oscillator after loosing synch */
  u_long cl_cflag;              /* terminal io flags */
  u_long cl_iflag;              /* terminal io flags */
  u_long cl_oflag;              /* terminal io flags */
  u_long cl_lflag;              /* terminal io flags */
} clockinfo[] =
{				/*   0.  0.0.128 - base offset for PPS support */
  {				/* 127.127.8.<device> */
    NO_POLL,
    NO_INIT,
    NO_END,
    NO_DATA,
    DCFPZF535_ROOTDELAY,
    DCFPZF535_BASEDELAY,
    DCFPZF535_PRECISION,
    DCF_ID,
    DCFPZF535_DESCRIPTION,
    NO_FORMAT,
    DCFPZF535_MAXUNSYNC,
    DCFPZF535_CFLAG,
    DCFPZF535_IFLAG,
    DCFPZF535_OFLAG,
    DCFPZF535_LFLAG
  },
  {				/* 127.127.8.16+<device> */
    NO_POLL,
    NO_INIT,
    NO_END,
    NO_DATA,
    DCFPZF535OCXO_ROOTDELAY,
    DCFPZF535OCXO_BASEDELAY,
    DCFPZF535OCXO_PRECISION,
    DCF_ID,
    DCFPZF535OCXO_DESCRIPTION,
    NO_FORMAT,
    DCFPZF535OCXO_MAXUNSYNC,
    DCFPZF535OCXO_CFLAG,
    DCFPZF535OCXO_IFLAG,
    DCFPZF535OCXO_OFLAG,
    DCFPZF535OCXO_LFLAG
  },
  {				/* 127.127.8.32+<device> */
    NO_POLL,
    NO_INIT,
    NO_END,
    NO_DATA,
    DCFUA31_ROOTDELAY,
    DCFUA31_BASEDELAY,
    DCFUA31_PRECISION,
    DCF_ID,
    DCFUA31_DESCRIPTION,
    NO_FORMAT,
    DCFUA31_MAXUNSYNC,
    DCFUA31_CFLAG,
    DCFUA31_IFLAG,
    DCFUA31_OFLAG,
    DCFUA31_LFLAG
  },
  {				/* 127.127.8.48+<device> */
    NO_POLL,
    NO_INIT,
    NO_END,
    NO_DATA,
    DCF7000_ROOTDELAY,
    DCF7000_BASEDELAY,
    DCF7000_PRECISION,
    DCF_ID,
    DCF7000_DESCRIPTION,
    NO_FORMAT,
    DCF7000_MAXUNSYNC,
    DCF7000_CFLAG,
    DCF7000_IFLAG,
    DCF7000_OFLAG,
    DCF7000_LFLAG
  }
};

int ncltypes = sizeof(clockinfo) / sizeof(struct clockinfo);

#define CL_TYPE(x)  (((((x) >> 4) & 0x7) > ncltypes) ? 0 : ((x) >> 4) & 0x7)
#define CL_PPS(x)   (x & 0x80)
#define CL_UNIT(x)  ((x) & 0xF)
  
/*
 * Other constant stuff
 */
#define	DCFHSREFID	0x7f7f08ff	/* 127.127.8.255 refid for hi strata */

#define DCFNOSYNCREPEAT (10*60)		/* mention uninitialized clocks all 10 minutes */
#define DCFSTATISTICS   (60*60)	        /* output state statistics every hour */

struct dcfunit
{
  /*
   * XNTP management
   */
  struct peer        *peer;		/* backlink to peer structure - refclock inactive if 0  */
  int                 fd;		/* device file descriptor */
  u_char              unit;		/* encoded unit/type/PPS */
  struct refclockio   io;		/* io system structure (used in PPS mode) */
#ifndef STREAM
  dcf_t		      dcfio;	        /* io handling structure */
#endif

  /*
   * type specific parameters
   */
  struct clockinfo   *dcf_type;	        /* link to clock description */

  /*
   * clock specific configuration
   */
  l_fp                basedelay;        /* clock local phase offset */

  /*
   * clock state handling/reporting
   */
  u_char	      flags;	        /* flags (leap_control) */
  u_char              status;		/* current status */
  u_char              lastevent; 	/* last not NORMAL status */
  u_long	      lastchange;       /* time (xntp) when last state change accured */
  u_long	      statetime[CEVNT_MAX+1]; /* accumulated time of clock states */
  struct event        stattimer;        /* statistics timer */
  u_long              polls;		/* polls from NTP protocol machine */
  u_long              noresponse; 	/* number of expected but not seen datagrams */
  u_long              badformat; 	/* bad format (failed format conversions) */
  u_long              baddata;		/* usually bad receive length, bad format */

  u_char              pollonly;		/* 1 for polling only (no PPS mode) */
  u_char              pollneeddata; 	/* 1 for receive sample expected in PPS mode */
  u_long              laststatus;       /* last packet status (error indication) */
  u_short	      lastformat;       /* last format used */
  u_long              lastsync;		/* time (xntp) when clock was last seen fully synchronized */
  u_long              timestarted; 	/* time (xntp) when peer clock was instantiated */
  u_long	      nosynctime; 	/* time (xntp) when last nosync message was posted */
  u_long              lastmissed;       /* time (xntp) when poll didn't get data (powerup heuristic) */
  dcftime_t           time;		/* last (dcf module) data */
  void               *localdata;        /* optional local data */
};

static struct dcfunit *dcfunits[MAXUNITS];

static void dcf_receive();
static void dcf_process();

extern int io_addclock();
extern void io_closeclock();

extern u_long current_time;
extern struct event timerqueue[];

extern u_long ustotslo[];
extern u_long ustotsmid[];
extern u_long ustotshi[];

extern int pps_sample();

#define DCF_STATETIME(dcf, i) ((dcf->status == i) ? dcf->statetime[i] + current_time - dcf->lastchange : dcf->statetime[i])
      
/*--------------------------------------------------
 * convert a flag field to a string
 */
static char *dcfstate(state, buffer)
  unsigned long state;
  char *buffer;
{
  static struct bits
    {
      unsigned long bit;
      char         *name;
    } flagstrings[] =
    {
      { DCFB_ANNOUNCE, "DST SWITCH WARNING" },
      { DCFB_POWERUP,  "NOT SYNCHRONIZED" },
      { DCFB_NOSYNC,   "TIME CODE NOT CONFIRMED" },
      { DCFB_DST,      "DST" },
      { DCFB_UTC,      "UTC DISPLAY" },
      { DCFB_LEAP,     "LEAP WARNING" },
      { DCFB_ALTERNATE,"ALTERNATE ANTENNA" },
      { DCFB_TIMECODE, "TIME CODE" },
      { DCFB_PPS,      "PPS" },
      { 0 }
    };

  static struct sbits
    {
      unsigned long bit;
      char         *name;
    } sflagstrings[] =
    {
      { DCFB_S_LEAP,     "LEAP INDICATION" },
      { DCFB_S_PPS,      "PPS SIGNAL" },
      { DCFB_S_LOCATION, "LOCATION" },
      { 0 }
    };
  int i;

  *buffer = '\0';
  
  i = 0;
  while (flagstrings[i].bit)
    {
      if (flagstrings[i].bit & state)
	{
	  if (buffer[0])
	    strcat(buffer, "; ");
	  strcat(buffer, flagstrings[i].name);
	}
      i++;
    }
  
  if (state & (DCFB_S_LEAP|DCFB_S_LOCATION|DCFB_S_PPS))
    {
      register char *s, *t;
      
      if (buffer[0])
	strcat(buffer, "; ");

      strcat(buffer, "(");

      t = s = buffer + strlen(buffer);
      
      i = 0;
      while (sflagstrings[i].bit)
	{
	  if (sflagstrings[i].bit & state)
	    {
	      if (t != s)
		{
		  strcpy(t, "; ");
		  t += 2;
		}
	      
	      strcpy(t, sflagstrings[i].name);
	      t += strlen(t);
	    }
	  i++;
	}
      strcpy(t, ")");
    }
  return buffer;
}

/*--------------------------------------------------
 * convert a status flag field to a string
 */
static char *dcfstatus(state, buffer)
  unsigned long state;
  char *buffer;
{
  static struct bits
    {
      unsigned long bit;
      char         *name;
    } flagstrings[] =
    {
      { CVT_OK,      "CONVERSION SUCCESSFUL" },
      { CVT_NONE,    "NO CONVERSION" },
      { CVT_FAIL,    "CONVERSION FAILED" },
      { CVT_BADFMT,  "ILLEGAL FORMAT" },
      { CVT_BADDATE, "DATE ILLEGAL" },
      { CVT_BADTIME, "TIME ILLEGAL" },
      { 0 }
    };
  int i;

  *buffer = '\0';
  
  i = 0;
  while (flagstrings[i].bit)
    {
      if (flagstrings[i].bit & state)
	{
	  if (buffer[0])
	    strcat(buffer, "; ");
	  strcat(buffer, flagstrings[i].name);
	}
      i++;
    }
  
  return buffer;
}

/*--------------------------------------------------
 * convert a clock status flag field to a string
 */
static char *clockstatus(state)
  unsigned long state;
{
  static char buffer[20];
  static struct status
    {
      unsigned long value;
      char         *name;
    } flagstrings[] =
    {
      { CEVNT_NOMINAL, "NOMINAL" },
      { CEVNT_TIMEOUT, "NO RESPONSE" },
      { CEVNT_BADREPLY,"BAD FORMAT" },
      { CEVNT_FAULT,   "FAULT" },
      { CEVNT_PROP,    "PROPAGATION DELAY" },
      { CEVNT_BADDATE, "ILLEGAL DATE" },
      { CEVNT_BADTIME, "ILLEGAL TIME" },
      { ~0 }
    };
  int i;

  i = 0;
  while (flagstrings[i].value != ~0)
    {
      if (flagstrings[i].value == state)
	{
	  return flagstrings[i].name;
	}
      i++;
    }

  sprintf(buffer, "unknown #%d", state);
  
  return buffer;
}

/*--------------------------------------------------
 * mkascii - make a printable ascii string
 * assumes (unless defined better) 7-bit ASCII
 */
#ifndef isprint
#define isprint(_X_) (((_X_) > 0x1F) && ((_X_) < 0x7F))
#endif

static char *mkascii(buffer, blen, src, srclen)
  register char  *buffer;
  register long  blen;
  register char  *src;
  register long  srclen;
{
  register char *b    = buffer;
  register char *endb = (char *)0;
  
  if (blen < 4)
    return (char *)0;		/* don't bother with mini buffers */
  
  endb = buffer + blen - 4;
      
  blen--;			/* account for '\0' */

  while (blen && srclen--)
    {
      if ((*src != '\\') && isprint(*src))
	{			/* printables are easy... */
	  *buffer++ = *src++;
	  blen--;
	}
      else
	{
	  if (blen < 4)
	    {
	      while (blen--)
		{
		  *buffer++ = '.';
		}
	      *buffer = '\0';
	      return b;
	    }
	  else
	    {
	      if (*src == '\\')
		{
		  strcpy(buffer,"\\\\");
		  buffer += 2;
		  blen   -= 2;
		}
	      else
		{
		  sprintf(buffer, "\\x%02x", *src++);
		  blen   -= 4;
		  buffer += 4;
		}
	    }
	}
      if (srclen && !blen && endb) /* overflow - set last chars to ... */
	strcpy(endb, "...");
    }

  *buffer = '\0';
  return b;
}
  

/*--------------------------------------------------
 * mktime - make representation of a relative time
 */
static char *mktime(delta)
  unsigned long delta;
{
  unsigned long tmp, m, s;
  static char buffer[40];
  
  buffer[0] = '\0';
  
  if (tmp = delta / (60*60*24))
    {
      sprintf(buffer, "%dd+", tmp);
      delta -= tmp * 60*60*24;
    }

  s = delta % 60;
  delta /= 60;
  m = delta % 60;
  delta /= 60;

  sprintf(buffer+strlen(buffer), "%02d:%02d:%02d",
	  delta, m, s);

  return buffer;
}

  
/*--------------------------------------------------
 * dcf_init - initialize internal dcf driver data
 */
void
dcf_init()
{
  bzero(dcfunits, sizeof dcfunits);
}

/*--------------------------------------------------
 * dcf_statistics - list summary of clock states
 */
static void dcf_statistics(dcf)
  register struct dcfunit *dcf;
{
  register int i;

  syslog(LOG_INFO, "DCF77 receiver #%d: running time: %s",
	 CL_UNIT(dcf->unit),
	 mktime(current_time - dcf->timestarted));

  syslog(LOG_INFO, "DCF77 receiver #%d: current status: %s",
	 CL_UNIT(dcf->unit),
	 clockstatus(dcf->status));
  
  for (i = 0; i <= CEVNT_MAX; i++)
    {
      register unsigned long stime;
      register unsigned long percent, div = current_time - dcf->timestarted;

      percent = stime = DCF_STATETIME(dcf, i);

      while (((unsigned long)(~0) / 10000) < percent)
	{
	  percent /= 10;
	  div     /= 10;
	}

      if (div)
	percent = (percent * 10000) / div;
      else
	percent = 10000;

      if (stime)
	syslog(LOG_INFO, "DCF77 receiver #%d: state %18s: %13s (%3d.%02d%%)",
	       CL_UNIT(dcf->unit),
	       clockstatus(i),
	       mktime(stime),
	       percent / 100, percent % 100);
    }
}

/*--------------------------------------------------
 * cdcf_statistics - wrapper for statistics call
 */
static void cdcf_statistics(peer)
  register struct peer *peer;
{
  register struct dcfunit *dcf = (struct dcfunit *)peer;
  
  dcf_statistics(dcf);
  dcf->stattimer.event_time    = current_time + DCFSTATISTICS;
  TIMER_ENQUEUE(timerqueue, &dcf->stattimer);
}

/*--------------------------------------------------
 * dcf_shutdown - shut down a DCF clock
 */
void
dcf_shutdown(unit)
	int unit;
{
	register struct dcfunit *dcf;

	unit = CL_UNIT(unit);
	
	if (unit >= MAXUNITS) {
		syslog(LOG_ERR,
		  "DCF77 receiver #%d: dcf_shutdown: INTERNAL ERROR, unit invalid (max %d)",
		    unit,MAXUNITS);
		return;
	}

	dcf = dcfunits[unit];
	
	if (dcf && !dcf->peer) {
		syslog(LOG_ERR,
		 "DCF77 receiver #%d: dcf_shutdown: INTERNAL ERROR, unit not in use", unit);
		return;
	}

	/*
	 * print statistics a last time and
	 * stop statistics machine
	 */
	dcf_statistics(dcf);
	TIMER_DEQUEUE(&dcf->stattimer);
	
	/*
	 * Tell the I/O module to turn us off.  We're history.
	 */
	if (!dcf->pollonly)
	  io_closeclock(&dcf->io);
	else
	  (void) close(dcf->fd);

#ifndef STREAM
	dcf_ioend(&dcf->dcfio);
#endif
	if (dcf->dcf_type->cl_end)
	  {
	    dcf->dcf_type->cl_end(dcf);
	  }
	
	syslog(LOG_INFO, "DCF77 receiver #%d: reference clock \"%s\" removed",
	       CL_UNIT(dcf->unit), dcf->dcf_type->cl_description);

	dcf->peer = (struct peer *)0; /* unused now */
}


/*--------------------------------------------------
 * dcf_start - open the DCF devices and initialize data for processing
 */
int
dcf_start(sysunit, peer)
	u_int sysunit;
	struct peer *peer;
{
  u_int unit;
  int fd232, i;
  struct termios tm;		/* NEEDED FOR A LONG TIME ! */
  struct strioctl strioc;
  struct dcfunit * dcf;
  dcfctl_t tmp_ctl;
  char dcfdev[sizeof(DCFDEVICE)+20];
  u_int type;

  type = CL_TYPE(sysunit);
  unit = CL_UNIT(sysunit);

  if (unit >= MAXUNITS)
    {
      syslog(LOG_ERR, "DCF77 receiver #%d: dcf_start: unit number invalid (max %d)",
	     unit, MAXUNITS);
      return 0;
    }

  if (dcfunits[unit] && dcfunits[unit]->peer)
    {
      syslog(LOG_ERR, "DCF77 receiver #%d: dcf_start: unit in use", unit);
      return 0;
    }

  /*
   * Unit okay, attempt to open the device.
   */
  (void) sprintf(dcfdev, DCFDEVICE, unit);

  fd232 = open(dcfdev, O_RDWR, 0777);
  if (fd232 == -1)
    {
      syslog(LOG_ERR, "DCF77 receiver #%d: dcf_start: open of %s failed: %m", unit, dcfdev);
      return 0;
    }

  /* 
   * configure terminal line
   */
  if (ioctl(fd232, TCGETS, &tm) == -1)
    {
      syslog(LOG_ERR, "DCF77 receiver #%d: dcf_start: ioctl(%d, TCGETS, &tm): %m", unit, fd232);
      close(fd232);
      return 0;
    }
  else
    {
      bzero(tm.c_cc, sizeof(tm.c_cc)); /* no specials - we'll do all ourself */

      tm.c_cc[VMIN]  = 1;	/* get any character */

      tm.c_cflag     = clockinfo[type].cl_cflag;
      tm.c_iflag     = clockinfo[type].cl_iflag;
      tm.c_oflag     = clockinfo[type].cl_oflag;
      tm.c_lflag     = clockinfo[type].cl_lflag;
	    
      if (ioctl(fd232, TCSETS, &tm) == -1)
	{
	  syslog(LOG_ERR, "DCF77 receiver #%d: dcf_start: ioctl(%d, TCSETS, &tm): %m", unit, fd232);
	  close(fd232);
	  return 0;
	}
    }

#ifdef I_POP
  /* 
   * pop all possibly active streams moduls from the tty
   */
  while(ioctl(fd232, I_POP, 0) == 0)
    /* empty loop !! */;
#endif

#ifdef STREAM
  /* 
   * now push the dcf streams module
   * it will ensure exclusive access to the device
   */
  if (ioctl(fd232, I_PUSH, "dcf") == -1)
    {
      syslog(LOG_ERR, "DCF77 receiver #%d: dcf_start: ioctl(%d, I_PUSH, \"dcf\"): %m",
	     unit, fd232);
      close(fd232);
      return 0;
    }
#else
  if (ioctl(fd232, TIOCEXCL, (char *)0) < 0)
    {
      syslog(LOG_ERR, "DCF77 receiver #%d: dcf_start: ioctl(%d, TIOCEXCL, 0) %m", unit,
	     fd232);
      close(fd232);
      return 0;
    }
#endif

  /*
   * Looks like this might succeed.  Find memory for the structure.
   * Look to see if there are any unused ones, if not we malloc()
   * one.
   */
  if (dcfunits[unit])
    {
      dcf = dcfunits[unit];	/* The one we want is okay - and free */
    }
  else
    {
      for (i = 0; i < MAXUNITS; i++)
	{
	  if (dcfunits[i] && !dcfunits[i]->peer)
	    break;
	}
      if (i < MAXUNITS)
	{
	  /*
	   * Reclaim this one
	   */
	  dcf = dcfunits[i];
	  dcfunits[i] = (struct dcfunit *)0;
	}
      else
	{
	  dcf = (struct dcfunit *)
	    emalloc(sizeof(struct dcfunit));
	}
    }

  bzero((char *)dcf, sizeof(struct dcfunit));
  dcfunits[unit] = dcf;

  /*
   * Set up the structures
   */
  dcf->unit         = (u_char)sysunit;
  dcf->timestarted  = current_time;
  dcf->lastchange   = current_time;
  /*
   * we want to filter input for the sake of
   * getting an impression on dispersion
   * also we like to average the median range
   */
  dcf->flags        = DCF_STAT_FILTER|DCF_STAT_AVG;
  dcf->pollneeddata = 0;
  dcf->pollonly     = 1;	/* go for default polling mode */
  dcf->lastformat   = ~0;	/* assume no format known */
  dcf->status	    = CEVNT_TIMEOUT; /* expect the worst */
  dcf->laststatus   = ~0;	/* be sure to mark initial status change */
  dcf->nosynctime   = 0;	/* assume clock reasonable */
  dcf->lastmissed   = 0;	/* assume got everything */
  dcf->localdata    = (void *)0;
  dcf->dcf_type     = &clockinfo[type];

  dcf->basedelay.l_ui = 0;	/* we can only pre-configure delays less than 1 second */
  dcf->basedelay.l_uf = dcf->dcf_type->cl_basedelay;


  if (dcf->dcf_type->cl_init)
    {
      if (dcf->dcf_type->cl_init(dcf))
	{
	  close(fd232);
	  return 0;		/* well, ok - active polling broke */
	}
    }
  
  /*
   * as we always(?) get 8 bit chars we want to be
   * sure, that the upper bits are zero for less
   * than 8 bit I/O - so we pass that information on.
   * note that there can be only one bit count format
   * per file descriptor
   */

  switch (tm.c_cflag & CSIZE)
    {
    case CS5:
      tmp_ctl.dcfsetcs.dcf_cs = DCF_IO_CS5;
      break;
      
    case CS6:
      tmp_ctl.dcfsetcs.dcf_cs = DCF_IO_CS6;
      break;
      
    case CS7:
      tmp_ctl.dcfsetcs.dcf_cs = DCF_IO_CS7;
      break;
      
    case CS8:
      tmp_ctl.dcfsetcs.dcf_cs = DCF_IO_CS8;
      break;
    }

#ifdef STREAM
  strioc.ic_cmd     = DCFIOC_SETCS;
  strioc.ic_timout  = 0;
  strioc.ic_dp      = (char *)&tmp_ctl;
  strioc.ic_len     = sizeof (tmp_ctl);
  
  if (ioctl(fd232, I_STR, &strioc) == -1)
    {
      syslog(LOG_ERR, "DCF77 receiver #%d: dcf_start: ioctl(%d, I_STR, DCFIOC_SETCS): %m", unit, fd232);
      close(fd232);
      return 0;
    }
#else
  if (!dcf_setcs(&tmp_ctl, dcf))
    {
      syslog(LOG_ERR, "DCF77 receiver #%d: dcf_start: dcf_setcs() FAILED.", unit);
      close(fd232);
      return 0;
    }
#endif  

  strcpy(tmp_ctl.dcfformat.dcf_buffer, dcf->dcf_type->cl_format);
  tmp_ctl.dcfformat.dcf_count = strlen(tmp_ctl.dcfformat.dcf_buffer);
  
#ifdef STREAM
  strioc.ic_cmd     = DCFIOC_SETFMT;
  strioc.ic_timout  = 0;
  strioc.ic_dp      = (char *)&tmp_ctl;
  strioc.ic_len     = sizeof (tmp_ctl);
  
  if (ioctl(fd232, I_STR, &strioc) == -1)
    {
      syslog(LOG_ERR, "DCF77 receiver #%d: dcf_start: ioctl(%d, I_STR, DCFIOC_SETFMT): %m", unit, fd232);
      close(fd232);
      return 0;
    }
#else
  if (!dcf_setfmt(&tmp_ctl, dcf))
    {
      syslog(LOG_ERR, "DCF77 receiver #%d: dcf_start: dcf_setfmt() FAILED.", unit);
      close(fd232);
      return 0;
    }
#endif  

  /*
   * get rid of all IO accumulated so far
   */
  {
    int flshcmd = TCIOFLUSH;

    (void) ioctl(fd232, TCFLSH, &flshcmd);
  }

  /* We do not want to get signals on any packet from the streams module
   * It will keep one old packet and after this is read the next read
   * will return an accurate new packet to be processed.
   */
  dcf->fd = fd232;
	
  /*
   * All done.  Initialize a few random peer variables, then
   * return success.
   */
  peer->rootdelay  = dcf->dcf_type->cl_rootdelay;
  peer->precision  = dcf->dcf_type->cl_precision;
  peer->stratum    = STRATUM_REFCLOCK;
  if (peer->stratum <= 1)
    bcopy(dcf->dcf_type->cl_id, (char *)&peer->refid, 4);
  else
    peer->refid = htonl(DCFHSREFID);
	
  /*
   * hot setup section
   * we need the asynch IO on non STREAMS operation and/or
   * when PPS is configured
   */
#if (defined(PPS) && defined(DCFPPS)) || !defined(STREAM)
#ifdef STREAM
  if (CL_PPS(dcf->unit))
#endif
    {
      /*
       * Insert in device list.
       */
      dcf->io.clock_recv = dcf_receive;
      dcf->io.srcclock = (caddr_t)dcf;
      dcf->io.datalen = 0;
      dcf->io.fd = dcf->fd;	/* replicated, but what the heck */
      if (!io_addclock(&dcf->io))
	{
#ifdef STREAM
	  syslog(LOG_ERR,
		 "DCF77 receiver #%d: dcf_start: addclock %s fails (switching to polling mode)", CL_UNIT(dcf->unit), dcfdev);
#else
	  syslog(LOG_ERR,
		 "DCF77 receiver #%d: dcf_start: addclock %s fails (ABORT)", CL_UNIT(dcf->unit), dcfdev);
	  dcf_shutdown(dcf->unit); /* let our cleaning staff do the work */
	  return 0;
#endif
	}
      else 
	{
	  dcf->pollonly = 0;	/*
				 * update at receipt of time_stamp - also
				 * supports PPS processing
				 */
	}
    }
#endif	

  dcf->peer = peer;		/* marks it also as busy */

  /*
   * wind up statistics timer
   */
  dcf->stattimer.peer = (struct peer *)dcf; /* we know better, but what the heck */
  dcf->stattimer.event_handler = cdcf_statistics;
  dcf->stattimer.event_time    = current_time + DCFSTATISTICS;
  TIMER_ENQUEUE(timerqueue, &dcf->stattimer);

  /*
   * print out configuration
   */
  syslog(LOG_INFO, "DCF77 receiver #%d: reference clock \"%s\" (device %s) added",
	 CL_UNIT(dcf->unit), 
	 dcf->dcf_type->cl_description, dcfdev);
  
  syslog(LOG_INFO, "DCF77 receiver #%d:  Stratum %d, %sPPS support, trust time %s, precision %d",
	 CL_UNIT(dcf->unit), 
	 dcf->peer->stratum, dcf->pollonly ? "no " : "",
	 mktime(dcf->dcf_type->cl_maxunsync), dcf->peer->precision);
  
  syslog(LOG_INFO, "DCF77 receiver #%d:  rootdelay %s s, phaseadjust %s s, %s IO handling",
	 CL_UNIT(dcf->unit), 
	 ufptoa(dcf->dcf_type->cl_rootdelay, 6),
	 lfptoa(dcf->basedelay, 8),
#ifdef STREAM
	 "STREAM"
#else
	 "normal"
#endif
	 );

  syslog(LOG_INFO, "DCF77 receiver #%d:  Format recognition: %s", CL_UNIT(dcf->unit),
	 !(*dcf->dcf_type->cl_format) ? "<AUTOMATIC>" : dcf->dcf_type->cl_format);
  
  return 1;
}

/*--------------------------------------------------
 * event handling - note that nominal events will also be posted
 */

static void dcf_event(dcf, event)
  struct dcfunit *dcf;
  int event;
{
  if (dcf->status != (u_char) event)
    {
      dcf->statetime[dcf->status] += current_time - dcf->lastchange;
      dcf->lastchange              = current_time;

      dcf->status    = (u_char)event;
      if (event != CEVNT_NOMINAL)
        dcf->lastevent = dcf->status;

      report_event(EVNT_PEERCLOCK, dcf->peer);
    }
}

/*--------------------------------------------------
 * dcf_receive - called by io handler
 *
 * this routine is called each time a time packed is
 * deliverd by the STREAMS module
 * if PPS processing is requested, pps_sample is called
 */
static void dcf_receive(rbufp)
  struct recvbuf *rbufp;
{
  struct dcfunit *dcf = (struct dcfunit *)rbufp->recv_srcclock;
  dcftime_t dcftime;
#ifndef STREAM
  register int count;
  register char *s;
#endif

#ifdef STREAM
  if (rbufp->recv_length != sizeof(dcftime_t))
    {
      syslog(LOG_ERR,"DCF77 receiver #%d: dcf_receive: bad size (got %d expected %d)",
	     CL_UNIT(dcf->unit), rbufp->recv_length, sizeof(dcftime_t));
      dcf->baddata++;
      dcf_event(dcf, CEVNT_BADREPLY);
      return;
    }
  bcopy((caddr_t)&rbufp->recv_space, (caddr_t)&dcftime, sizeof(dcftime_t));

  dcf_process(dcf, &dcftime);
#else
  /*
   * eat all characters, parsing then and feeding complete samples
   */
  count = rbufp->recv_length;
  s = rbufp->recv_buffer;
  
  while (count--)
    {
      int cvtrtc;
      
      if (dcf_ioread(&dcf->dcfio, *s++, &rbufp->recv_time, &dcftime))
	{
	  /*
	   * got something good to eat
	   */
	  dcf_process(dcf, &dcftime);
	  dcf_iodone(&dcf_dcfio);
	}
    }
#endif
}

/*--------------------------------------------------
 * dcf_poll - called by the transmit procedure
 */
void
dcf_poll(unit, peer)
	int unit;
	struct peer *peer;
{ 
  extern int errno;

  register int fd, i, rtc;
  fd_set fdmask;
  struct timeval timeout;
  register struct dcfunit *dcf;
  dcftime_t dcftime;

  unit = CL_UNIT(unit);
  
  if (unit >= MAXUNITS)
    {
      syslog(LOG_ERR, "DCF77 receiver #%d: poll: INTERNAL: unit invalid",
	     unit);
      return;
    }

  dcf = dcfunits[unit];

  if (!dcf->peer)
    {
      syslog(LOG_ERR, "DCF77 receiver #%d: poll: INTERNAL: unit unused",
	     unit);
      return;
    }

  if (peer != dcf->peer)
    {
      syslog(LOG_ERR,
	     "DCF77 receiver #%d: poll: INTERNAL: peer incorrect",
	     unit);
      return;
    }

  /*
   * Update clock stat counters
   */
  dcf->polls++;

#if (defined(PPS) && defined(DCFPPS)) || !defined(STREAM)
  /*
   * in PPS mode we just mark that we want the next sample
   * for the clock filter
   */
#ifdef STREAM
  if (!dcf->pollonly)
#endif
    {
      if (dcf->pollneeddata)
	{
	  /*
	   * bad news - didn't get a response last time
	   */
	  dcf->noresponse++;
	  dcf->lastmissed = current_time;
	  dcf_event(dcf, CEVNT_TIMEOUT);

          syslog(LOG_WARNING, "DCF77 receiver #%d: no data from device within poll interval", CL_UNIT(dcf->unit));
	}
      dcf->pollneeddata = 1;
      return;
    }
#endif

  if (dcf->dcf_type->cl_poll)
    {
      dcf->dcf_type->cl_poll(dcf);
    }
  
#ifdef STREAM
  /* 
   * now we do the following:
   *    - read the first packet from the dcf module  (OLD !!!)
   *    - read the second packet from the dcf module (fresh)
   *    - compute values for xntp
   */
	  
  FD_ZERO(&fdmask);
  fd = dcf->fd;
  FD_SET(fd, &fdmask);
  timeout.tv_sec = 0;
  timeout.tv_usec = 500000;	/* 0.5 sec */

  while ((rtc = select(fd + 1, &fdmask, 0, 0, &timeout)) != 1)
    {
      /* no data from the radio clock */

      if (rtc == -1)
	{
	  if (errno == EINTR)
	    {
	      continue;
	    }
	  else
	    {
              syslog(LOG_WARNING, "DCF77 receiver #%d: no data[old] from device (select() error: %m)", unit);
	    }
	}
      else
	{
          syslog(LOG_WARNING, "DCF77 receiver #%d: no data[old] from device", unit);
	}
      dcf->noresponse++;
      dcf->lastmissed = current_time;
      dcf_event(dcf, CEVNT_TIMEOUT);
      
      return;
    }

  while (((i = read(fd, &dcftime, sizeof(dcftime))) < sizeof(dcftime)))
    {
      /* bad packet */
      if ( i == -1)
	{
	  if (errno == EINTR)
	    {
	      continue;
	    }
	  else
	    {
              syslog(LOG_WARNING, "DCF77 receiver #%d: bad read[old] from streams module (read() error: %m)", unit, i, sizeof(dcftime));
	    }
	}
      else
	{
          syslog(LOG_WARNING, "DCF77 receiver #%d: bad read[old] from streams module (got %d bytes - expected %d bytes)", unit, i, sizeof(dcftime));
	}
      dcf->baddata++;
      dcf_event(dcf, CEVNT_BADREPLY);
      
      return;
    }

  timeout.tv_sec = 1;
  timeout.tv_usec = 500000;	/* 1.500 sec */
  FD_ZERO(&fdmask);
  FD_SET(fd, &fdmask);
  while ((rtc = select(fd + 1, &fdmask, 0, 0, &timeout)) != 1)
    {
      /* no data from the radio clock */

      if (rtc == -1)
	{
	  if (errno == EINTR)
	    {
	      continue;
	    }
	  else
	    {
              syslog(LOG_WARNING, "DCF77 receiver #%d: no data[new] from device (select() error: %m)", unit);
	    }
	}
      else
	{
          syslog(LOG_WARNING, "DCF77 receiver #%d: no data[new] from device", unit);
	}
	      
      dcf->noresponse++;
      dcf->lastmissed = current_time;
      dcf_event(dcf, CEVNT_TIMEOUT);
      
      return;
    }

  if (((i = read(fd, &dcftime, sizeof(dcftime))) < sizeof(dcftime)))
    {
      /* bad packet */
      if ( i== -1)
	{
          syslog(LOG_WARNING, "DCF77 receiver #%d: bad read[new] from streams module (read() error: %m)", unit, i, sizeof(dcftime));
	}
      else
	{
          syslog(LOG_WARNING, "DCF77 receiver #%d: bad read[new] from streams module (got %d bytes - expected %d bytes)", unit, i, sizeof(dcftime));
	}
      dcf->baddata++;
      dcf_event(dcf, CEVNT_BADREPLY);
      
      return;
    }
  dcf_process(dcf, &dcftime);
#endif
}

/*--------------------------------------------------
 * process a DCF time sample
 */
static void dcf_process(dcf, dcftime)
  struct dcfunit *dcf;
  dcftime_t      *dcftime;
{
  struct timeval usecdisp;
  l_fp off, rectime, reftime, dispersion;
  
  /*
   * check for changes in conversion status
   * (only one for each new status !)
   */
  if (dcf->laststatus != dcftime->dcf_status)
    {
      char buffer[200];

      syslog(LOG_WARNING, "DCF77 receiver #%d: conversion status \"%s\"",
	     CL_UNIT(dcf->unit), dcfstatus(dcftime->dcf_status, buffer));

      if ((dcftime->dcf_status & CVT_MASK) == CVT_FAIL)
	{
	  /*
	   * tell more about the story - list time code
	   * there is a slight change for a race condition and
	   * the time code might be overwritten by the next packet
	   */
	  dcfctl_t tmpctl;
#ifdef STREAM
	  struct strioctl strioc;

	  strioc.ic_cmd     = DCFIOC_TIMECODE;
	  strioc.ic_timout  = 0;
	  strioc.ic_dp      = (char *)&tmpctl;
	  strioc.ic_len     = sizeof (tmpctl);
	
	  if (ioctl(dcf->fd, I_STR, &strioc) == -1)
	    {
	      syslog(LOG_ERR, "DCF77 receiver #%d: dcf_process: ioctl(%d, I_STR, DCFIOC_TIMECODE): %m", CL_UNIT(dcf->unit), dcf->fd);
	    }
#else
	  if (!dcf_timecode(&tmpctl, &dcf->dcfio))
	    {
	      syslog(LOG_ERR, "DCF77 receiver #%d: dcf_process: dcf_timecode() FAILED", CL_UNIT(dcf->unit));
	    }
#endif	      
	  else
	    {
	      syslog(LOG_WARNING, "DCF77 receiver #%d: FAILED TIMECODE: \"%s\"",
		     CL_UNIT(dcf->unit), mkascii(buffer, sizeof buffer, tmpctl.dcfgettc.dcf_buffer, tmpctl.dcfgettc.dcf_count - 1));
	      dcf->badformat += tmpctl.dcfgettc.dcf_badformat;
	    }
	}
      
      dcf->laststatus = dcftime->dcf_status;
    }

  /*
   * examine status and post appropriate events
   */
  if ((dcftime->dcf_status & CVT_MASK) != CVT_OK)
    {
      /*
       * got bad data - tell the rest of the system
       */
      switch (dcftime->dcf_status & CVT_MASK)
	{
	case CVT_NONE:
	  break;		/* well, still waiting - timeout is handled at higher levels */

	case CVT_FAIL:
	  dcf->badformat++;
	  if (dcftime->dcf_status & CVT_BADFMT)
	    {
	      dcf_event(dcf, CEVNT_BADREPLY);
	    }
	  else
	    if (dcftime->dcf_status & CVT_BADDATE)
	      {
		dcf_event(dcf, CEVNT_BADDATE);
	      }
	    else
	      if (dcftime->dcf_status & CVT_BADTIME)
		{
		  dcf_event(dcf, CEVNT_BADTIME);
		}
	      else
		{
		  dcf_event(dcf, CEVNT_BADREPLY); /* for the lack of something better */
		}
	}
      return;			/* skip the rest - useless */
    }

  /*
   * check for format changes
   * (in case somebody has swapped clocks 8-)
   */
  if (dcf->lastformat != dcftime->dcf_format)
    {
      dcfctl_t tmpctl;
#ifdef STREAM
      struct strioctl strioc;

      strioc.ic_cmd     = DCFIOC_GETFMT;
      strioc.ic_timout  = 0;
      strioc.ic_dp      = (char *)&tmpctl;
      strioc.ic_len     = sizeof (tmpctl);
#endif
	
      tmpctl.dcfformat.dcf_format = dcftime->dcf_format;
      
#ifdef STREAM
      if (ioctl(dcf->fd, I_STR, &strioc) == -1)
	{
	  syslog(LOG_ERR, "DCF77 receiver #%d: ioctl(%d, I_STR, DCFIOC_GETFMT): %m", CL_UNIT(dcf->unit), dcf->fd);
	}
#else
      if (!dcf_getfmt(&tmpctl, &dcf->dcfio))
	{
	  syslog(LOG_ERR, "DCF77 receiver #%d: dcf_getfmt() FAILED", CL_UNIT(dcf->unit));
	}
#endif
      else
	{
	  syslog(LOG_INFO, "DCF77 receiver #%d: new packet format \"%s\"",
		 CL_UNIT(dcf->unit), tmpctl.dcfformat.dcf_buffer);
	}
      dcf->lastformat = dcftime->dcf_format;
    }
  
  /*
   * now, any changes ?
   */
  if (dcf->time.dcf_state != dcftime->dcf_state)
    {
      char tmp1[200];
      char tmp2[200];
      /*
       * something happend
       */
	    
      (void) dcfstate(dcftime->dcf_state, tmp1);
      (void) dcfstate(dcf->time.dcf_state, tmp2);
	    
      syslog(LOG_INFO,"DCF77 receiver #%d: STATE CHANGE: %s -> %s",
	     CL_UNIT(dcf->unit), tmp2, tmp1);
    }

  /*
   * remember for future
   */
  dcf->time = *dcftime;

  /*
   * check to see, whether the clock did a complete powerup or lost PZF signal
   * and post correct events for current condition
   */
  if (DCF_POWERUP(dcftime->dcf_state))
    {
      /*
       * this is bad, as we have completely lost synchronisation
       * well this is a problem with the receiver here
       * for DCF U/A 31 the lost synchronisation ist true
       * as it is the powerup state and the time is taken
       * from a crude real time clock chip
       * for the PZF series this is only partly true, as
       * DCF_POWERUP only means that the pseudo random
       * phase shift sequence cannot be found. this is only
       * bad, if we have never seen the clock in the SYNC
       * state, where the PHASE and EPOCH are correct.
       * for reporting events the above business does not
       * really matter, but we can use the time code
       * even in the POWERUP state after having seen
       * the clock in the synchronized state (PZF class
       * receivers) unless we have had a telegram disruption
       * after having seen the clock in the SYNC state. we
       * thus require having seen the clock in SYNC state
       * *after* having missed telegrams (noresponse) from
       * the clock. one problem remains: we might use erroneously
       * POWERUP data if the disruption is shorter than 1 polling
       * interval. fortunately powerdowns last usually longer than 64
       * seconds and the receiver is at least 2 minutes in the
       * POWERUP or NOSYNC state before switching to SYNC
       */
      dcf_event(dcf, CEVNT_FAULT);
      if (dcf->nosynctime)
	{
	  /*
	   * repeated POWERUP/NOSYNC state - look whether
	   * the message should be repeated
	   */
	  if (current_time - dcf->nosynctime > DCFNOSYNCREPEAT)
	    {
	      syslog(LOG_ERR,"DCF77 receiver #%d: *STILL* NOT SYNCHRONIZED (POWERUP or no PZF signal)",
		     CL_UNIT(dcf->unit));
	      dcf->nosynctime = current_time;
	    }
	}
      else
	{
	  syslog(LOG_ERR,"DCF77 receiver #%d: NOT SYNCHRONIZED (POWERUP or no PZF signal)",
		 CL_UNIT(dcf->unit));
          dcf->nosynctime = current_time;
	}
    }
  else
    {
      /*
       * we have two states left
       *
       * SYNC:
       *  this state means that the EPOCH (timecode) and PHASE
       *  information has be read correctly (at least two
       *  successive DCF77 timecodes were received correctly)
       *  this is the best possible state - full trust
       *
       * NOSYNC:
       *  The clock should be on phase with respect to the second
       *  signal, but the timecode has not been received correctly within
       *  at least the last two minutes. this is a sort of half baked state
       *  for DCF U/A 31 this is bad news (clock running without timecode
       *  confirmation)
       *  PZF 535 has also no time confirmation, but the phase should be
       *  very precise as the PZF signal can be decoded
       */
      dcf->nosynctime = 0;	/* current state is better than worst state */

      if (DCF_SYNC(dcftime->dcf_state))
	{
	  /*
	   * currently completely synchronized - best possible state
	   */
	  dcf->lastsync = current_time;
	  /*
	   * log OK status
	   */
	  dcf_event(dcf, CEVNT_NOMINAL);
	}
      else
	{
	  /*
	   * we have had some problems receiving the time code
	   */
	  dcf_event(dcf, CEVNT_PROP);
	}
    }

  if (DCF_PPS(dcftime->dcf_state))
    {
      l_fp offset;
      
      /*
       * we have a PPS signal - much better than the RS232 stuff
       */
      if (!buftvtots(&dcftime->dcf_ptime, &offset))
	return;

      if (DCF_TIMECODE(dcftime->dcf_state))
	{
	  if (buftvtots(&dcftime->dcf_time, &off))
	    {
	      /*
	       * WARNING: assumes on TIMECODE == PULSE (timecode after pulse)
	       */
	      rectime = reftime = off; /* take reference time - fake rectime */
	      L_SUB(&off, &offset); /* true offset */
	    }
	  else
	    {
	      /*
	       * assumes on second pulse
	       */
	      reftime = offset;	/* we don't have anything better */
	      rectime = offset;	/* to pass sanity checks */
	      off = offset;
	      off.l_ui = 0;
	      M_NEG(off.l_ui, off.l_f);	/* implied on second offset */
	    }
	}
      else
	{
	  /*
	   * assumes on second pulse
	   */
	  reftime = off = offset;
	  rectime = offset;
	  off.l_ui = 0;
	  M_NEG(off.l_ui, off.l_f); /* implied on second offset */
	}
    }
  else
    if (DCF_TIMECODE(dcftime->dcf_state))
      {
	l_fp ts;
	l_fp offset;
    
	/*
	 * calculate time offset including systematic delays
	 * off = DCF-timestamp + propagation delay - kernel time stamp
	 */
	offset = dcf->basedelay;
    
	if (!buftvtots(&dcftime->dcf_time, &off))
	  return;

	reftime = off;
    
	L_ADD(&off, &offset);
	rectime = off;		/* this makes org time and xmt time somewhat artificial */
    
#ifdef STREAM
	if (!buftvtots(&dcftime->dcf_stime, &ts))
	  return;
	L_SUB(&off, &ts);
#else
	L_SUB(&off, &dcftime->dcf_stime);
#endif
      }
    else
      {
	/*
	 * Well, no PPS, no TIMECODE, no work ...
	 */
	return;
      }
  
  
#if defined(PPS) && defined(DCFPPS)
  if (!dcf->pollonly && DCF_SYNC(dcftime->dcf_state))
    {
      /*
       * only provide PPS information when clock
       * is in sync
       * thus PHASE and EPOCH are correct
       */
      (void) pps_sample(&off);
    }

  /*
   * ready, unless the machine wants a sample
   */
  if (!dcf->pollonly && !dcf->pollneeddata)
    return;

  dcf->pollneeddata = 0;
#endif

  if (DCF_PPS(dcftime->dcf_state))
    {
      L_CLR(&dispersion);
    }
  else
    {
      /*
       * convert usec dispersion into NTP TS world
       */
  
      usecdisp.tv_sec  = dcftime->dcf_usecdisp / 1000000;
      usecdisp.tv_usec = dcftime->dcf_usecdisp % 1000000;

      TVTOTS(&usecdisp, &dispersion);
    }
  
  /*
   * and now stick it into the clock machine
   * samples are only valid iff lastsync is not too old and
   * we have seen the clock in sync at least once
   * after the last time we didn't see an expected data telegram
   * see the clock states section above for more reasoning
   */
  refclock_receive(dcf->peer, &off, 0, LFPTOFP(&dispersion), &reftime, &rectime,
		   (((current_time - dcf->lastsync) > dcf->dcf_type->cl_maxunsync) ||
		    (dcf->lastsync <= dcf->lastmissed)) ? LEAP_NOTINSYNC :
		   (DCF_LEAP(dcftime->dcf_state) ? ((dcf->flags & DCF_LEAP_DELETE) ? LEAP_DELSECOND : LEAP_ADDSECOND) : LEAP_NOWARNING));
}

/*--------------------------------------------------
 * dcf_leap - called when a leap second occurs
 */

dcf_leap()
{
	/*
	 * DCF77 does encode a leap warning... we are aware but not afraid of that
	 * as long as we get a little help for the direction from the operator until
	 * DCF77 encodes the LEAP correction direction.
	 */
}


/*--------------------------------------------------
 * dcf_control - set fudge factors, return statistics
 */
void
dcf_control(unit, in, out)
  u_int unit;
  struct refclockstat *in;
  struct refclockstat *out;
{
  register struct dcfunit *dcf;
  dcfctl_t tmpctl;
#ifdef STREAM
  struct strioctl strioc;
#endif  
  unsigned long type;
  static char outstatus[400];	/* status output buffer / hope this will be mutex */
  
  type = CL_TYPE(unit);
  unit = CL_UNIT(unit);
  
  if (out)
    {
      out->lencode       = 0;
      out->lastcode      = 0;
      out->polls         = out->noresponse = 0;
      out->badformat     = out->baddata    = 0;
      out->timereset     = 0;
      out->currentstatus = out->lastevent = CEVNT_NOMINAL;
    }

  if (unit >= MAXUNITS)
    {
      syslog(LOG_ERR, "DCF77 receiver #%d: dcf_control: unit invalid (max %d)",
	     unit, MAXUNITS-1);
      return;
    }

  dcf = dcfunits[unit];

  if (!dcf || !dcf->peer)
    {
      syslog(LOG_ERR, "DCF77 receiver #%d: dcf_control: unit invalid (UNIT INACTIVE)",
	     unit);
      return;
    }

  if (in)
    {
      if (in->haveflags & CLK_HAVETIME1)
	dcf->basedelay = in->fudgetime1;

      if (in->haveflags & CLK_HAVETIME2)
	{
	  /* not USED */
	}
      
      if (in->haveflags & CLK_HAVEVAL1)
	{
	  dcf->peer->stratum = (u_char)(in->fudgeval1 & 0xf);
	  if (dcf->peer->stratum <= 1)
		bcopy(dcf->dcf_type->cl_id, (char *)&dcf->peer->refid, 4);
	      else
		dcf->peer->refid = htonl(DCFHSREFID);
	}

      /*
       * NOT USED - yet
       *
      if (in->haveflags & CLK_HAVEVAL2)
	{
	}
       */
      if (in->haveflags & (CLK_HAVEFLAG1|CLK_HAVEFLAG2|CLK_HAVEFLAG3|CLK_HAVEFLAG4))
	{
	  dcf->flags = in->flags & (CLK_FLAG1|CLK_FLAG2|CLK_FLAG3|CLK_FLAG4);
	}

      if (in->haveflags & (CLK_HAVEVAL2|CLK_HAVETIME2|CLK_HAVEFLAG1|CLK_HAVEFLAG2|CLK_HAVEFLAG3|CLK_HAVEFLAG4))
	{
	  dcfctl_t tmpctl;
#ifdef STREAM
	  /* 
	   * now configure the streams module
	   */
	  strioc.ic_cmd     = DCFIOC_SETSTAT;
	  strioc.ic_timout  = 0;
	  strioc.ic_dp      = (char *)&tmpctl;
	  strioc.ic_len     = sizeof tmpctl;

	  tmpctl.dcfstatus.flags = dcf->flags & DCF_STAT_FLAGS;
	  
	if (ioctl(dcfunits[unit]->fd, I_STR, &strioc) == -1)
	  {
	    syslog(LOG_ERR, "DCF77 receiver #%d: dcf_control: ioctl(%d, I_STR, DCFIOC_SETSTAT): %m", unit, dcfunits[unit]->fd);
	  }
#else
        if (!dcf_setstat(&tmpctl, &dcf->dcfio))
	  {
	    syslog(LOG_ERR, "DCF77 receiver #%d: dcf_control: dcf_setstat() FAILED", unit);
	  }
#endif
	}
    }
  
  if (out)
    {
      register unsigned long sum = 0;
      register char *t;
      register struct tm *tm;
      register struct tm lctm;
      register short utcoff;
      register char sign;
      register int i;

      out->type = REFCLK_DCF77;
      out->haveflags = CLK_HAVETIME1|CLK_HAVEVAL1|CLK_HAVEFLAG1|CLK_HAVEFLAG2|CLK_HAVEFLAG3;
      out->clockdesc = dcf->dcf_type->cl_description;

      out->fudgetime1 = dcf->basedelay;
      
      out->fudgeval1 = (long)dcf->peer->stratum;

      out->flags     = dcf->flags;

      /*
       * all this for just finding out the +-xxxx part (there are always
       * new and changing fields in the standards 8-().
       *
       * but we do it for the human user...
       */
      tm = gmtime(&dcf->time.dcf_time.tv_sec);
      utcoff = tm->tm_hour * 60 + tm->tm_min;
      tm = localtime(&dcf->time.dcf_time.tv_sec);
      utcoff = tm->tm_hour * 60 + tm->tm_min - utcoff + 720;
      utcoff %= 1440;
      utcoff -= 720;
      if (utcoff < 0)
	{
	  utcoff = -utcoff;
	  sign = '-';
	}
      else
	{
	  sign = '+';
	}

      strcpy(outstatus, ctime(&dcf->time.dcf_time.tv_sec));
      t = rindex(outstatus, '\n');
      if (!t)
	{
	  t = outstatus + strlen(outstatus);
    	}
      else
	{
	  sprintf(t, " %c%02d%02d", sign, utcoff / 60, utcoff % 60);
	  t += strlen(t);
	}

#ifdef STREAM
      strioc.ic_cmd     = DCFIOC_TIMECODE;
      strioc.ic_timout  = 0;
      strioc.ic_dp      = (char *)&tmpctl;
      strioc.ic_len     = sizeof (tmpctl);
	
       if (ioctl(dcfunits[unit]->fd, I_STR, &strioc) == -1)
	{
	  syslog(LOG_ERR, "DCF77 receiver #%d: dcf_control: ioctl(%d, I_STR, DCFIOC_TIMECODE): %m", unit, dcfunits[unit]->fd);
	}
#else
      if (!dcf_timecode(&tmpctl, &dcf->dcfio))
	{
	  syslog(LOG_ERR, "DCF77 receiver #%d: dcf_control: dcf_timecode() FAILED", unit);
	}
#endif	      
      else
	{
	  if (t)
	    {
	      *t = ' ';
	      (void) dcfstate(tmpctl.dcfgettc.dcf_state, t+1);
	    }
	  else
	    {
	      strcat(outstatus, " ");
	      (void) dcfstate(tmpctl.dcfgettc.dcf_state, outstatus + strlen(outstatus));
	    }
	  strcat(outstatus," <");
	  mkascii(outstatus+strlen(outstatus), sizeof(outstatus) - strlen(outstatus) - 1,
		  tmpctl.dcfgettc.dcf_buffer, tmpctl.dcfgettc.dcf_count - 1);
	  strcat(outstatus,">");
	  dcf->badformat += tmpctl.dcfgettc.dcf_badformat;
	}
	  
      tmpctl.dcfformat.dcf_format = tmpctl.dcfgettc.dcf_format;
	  
#ifdef STREAM
      strioc.ic_cmd     = DCFIOC_GETFMT;

      if (ioctl(dcfunits[unit]->fd, I_STR, &strioc) == -1)
	{
	  syslog(LOG_ERR, "DCF77 receiver #%d: dcf_control: ioctl(%d, I_STR, DCFIOC_GETFMT): %m", unit, dcfunits[unit]->fd);
	}
#else
      if (!dcf_getfmt(&tmpctl, &dcf->dcfio))
	{
	  syslog (LOG_ERR, "DCF77 receiver #%d: dcf_control: dcf_getfmt() FAILED", unit);
	}
#endif
      else
	{
	  strcat(outstatus," (");
	  strncat(outstatus, tmpctl.dcfformat.dcf_buffer, tmpctl.dcfformat.dcf_count);
	  strcat(outstatus,")");
	}

      /*
       * gather state statistics
       */

      t = outstatus + strlen(outstatus);
      
      for (i = 0; i <= CEVNT_MAX; i++)
	{
	  register unsigned long stime;
	  register unsigned long div = current_time - dcf->timestarted;
	  register unsigned long percent;
	  
	  percent = stime = DCF_STATETIME(dcf, i);
      
	  while (((unsigned long)(~0) / 10000) < percent)
	    {
	      percent /= 10;
	      div     /= 10;
	    }
	  
	  if (div)
	    percent = (percent * 10000) / div;
	  else
	    percent = 10000;

	  if (stime)
	    {
	      sprintf(t, "%s%s: %s (%d.%02d%%)",
		      sum ? "; " : " [",
		      clockstatus(i),
		      mktime(stime),
		      percent / 100, percent % 100);
	      sum += stime;
	      t   += strlen(t);
	    }
	}

      sprintf(t, "; running time: %s]", mktime(sum));
      
      out->lencode       = strlen(outstatus);
      out->lastcode      = outstatus;
      out->timereset     = dcf->timestarted;
      out->polls         = dcf->polls;
      out->noresponse    = dcf->noresponse;
      out->badformat     = dcf->badformat;
      out->baddata       = dcf->baddata;
      out->lastevent     = dcf->lastevent;
      out->currentstatus = dcf->status;
    }
}
#endif
